"""Test the DAP handler, which forms the core of the client."""

import unittest

import numpy as np

import pydap.model
from pydap.handlers.dap import (
    BaseProxyDap2,
    DAPHandler,
    SequenceProxy,
    find_pattern_in_string_iter,
    walk,
)
from pydap.handlers.lib import BaseHandler, ConstraintExpression
from pydap.model import BaseType, DatasetType, GridType, StructureType
from pydap.tests.datasets import (
    SimpleArray,
    SimpleGrid,
    SimpleGroup,
    SimpleSequence,
    VerySimpleSequence,
)

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class TestDapHandler(unittest.TestCase):
    """Test that the handler creates the correct dataset from a URL."""

    def setUp(self):
        """Create WSGI apps"""
        self.app1 = BaseHandler(SimpleGrid)
        self.app2 = BaseHandler(SimpleSequence)
        self.app3 = BaseHandler(SimpleGrid, gzip=True)
        self.app4 = BaseHandler(SimpleGroup)

    def test_grid(self):
        """Test that dataset has the correct data proxies for grids."""
        dataset = DAPHandler("http://localhost:8001/", self.app1).dataset

        self.assertEqual(list(dataset.keys()), ["SimpleGrid", "x", "y"])
        self.assertEqual(list(dataset.SimpleGrid.keys()), ["SimpleGrid", "x", "y"])

        # test one of the axis
        self.assertIsInstance(dataset.SimpleGrid.x.data, BaseProxyDap2)
        self.assertEqual(dataset.SimpleGrid.x.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.SimpleGrid.x.data.id, "SimpleGrid.x")
        self.assertEqual(dataset.SimpleGrid.x.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.SimpleGrid.x.data.shape, (3,))
        self.assertEqual(dataset.SimpleGrid.x.data.slice, (slice(None),))

        # test the grid
        self.assertIsInstance(dataset.SimpleGrid.SimpleGrid.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.baseurl, "http://localhost:8001/"
        )
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.id, "SimpleGrid.SimpleGrid")
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (2, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice, (slice(None), slice(None))
        )
        self.assertEqual(
            repr(dataset.SimpleGrid[:]),
            "<GridType with array 'SimpleGrid' and maps 'x', 'y'>",
        )

    def test_grid_gzip(self):
        """Test that dataset has the correct data proxies for grids."""
        dataset = DAPHandler("http://localhost:8001/", self.app3).dataset

        self.assertEqual(list(dataset.keys()), ["SimpleGrid", "x", "y"])
        self.assertEqual(list(dataset.SimpleGrid.keys()), ["SimpleGrid", "x", "y"])

        # test one of the axis
        self.assertIsInstance(dataset.SimpleGrid.x.data, BaseProxyDap2)
        self.assertEqual(dataset.SimpleGrid.x.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.SimpleGrid.x.data.id, "SimpleGrid.x")
        self.assertEqual(dataset.SimpleGrid.x.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.SimpleGrid.x.data.shape, (3,))
        self.assertEqual(dataset.SimpleGrid.x.data.slice, (slice(None),))

        # test the grid
        self.assertIsInstance(dataset.SimpleGrid.SimpleGrid.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.baseurl, "http://localhost:8001/"
        )
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.id, "SimpleGrid.SimpleGrid")
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (2, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice, (slice(None), slice(None))
        )
        self.assertEqual(
            repr(dataset.SimpleGrid[:]),
            "<GridType with array 'SimpleGrid' and maps 'x', 'y'>",
        )

    def test_grid_erddap(self):
        """Test that dataset has the correct data proxies for grids
        with the ERDDAP behavior."""
        with patch(
            "pydap.handlers.lib.degenerate_grid_to_structure", side_effect=(lambda x: x)
        ) as mock_degenerate:
            dataset = DAPHandler("http://localhost:8001/", self.app1).dataset
            self.assertEqual(
                repr(dataset.SimpleGrid[:]),
                "<GridType with array 'SimpleGrid' and maps 'x', 'y'>",
            )
            assert mock_degenerate.called

    def test_grid_output_grid_false(self):
        """Test that dataset has the correct data proxies for grids with
        option output_grid set to False."""
        dataset = DAPHandler(
            "http://localhost:8001/", self.app1, output_grid=False
        ).dataset

        self.assertEqual(list(dataset.keys()), ["SimpleGrid", "x", "y"])
        self.assertEqual(list(dataset.SimpleGrid.keys()), ["SimpleGrid", "x", "y"])

        # test one of the axis
        self.assertIsInstance(dataset.SimpleGrid.x.data, BaseProxyDap2)
        self.assertEqual(dataset.SimpleGrid.x.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.SimpleGrid.x.data.id, "SimpleGrid.x")
        self.assertEqual(dataset.SimpleGrid.x.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.SimpleGrid.x.data.shape, (3,))
        self.assertEqual(dataset.SimpleGrid.x.data.slice, (slice(None),))

        # test the grid
        self.assertIsInstance(dataset.SimpleGrid.SimpleGrid.data, BaseProxyDap2)
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.baseurl, "http://localhost:8001/"
        )
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.id, "SimpleGrid.SimpleGrid")
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (2, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice, (slice(None), slice(None))
        )
        np.testing.assert_array_equal(dataset.SimpleGrid[:], [[0, 1, 2], [3, 4, 5]])

    def test_grid_erddap_output_grid_false(self):
        """Test that dataset has the correct data proxies for grids with
        option output_grid set to False and with the ERDDAP behavior."""
        with patch(
            "pydap.handlers.lib.degenerate_grid_to_structure", side_effect=(lambda x: x)
        ) as mock_degenerate:
            dataset = DAPHandler(
                "http://localhost:8001/", self.app1, output_grid=False
            ).dataset
            np.testing.assert_array_equal(dataset.SimpleGrid[:], [[0, 1, 2], [3, 4, 5]])
            assert mock_degenerate.called

    def test_grid_with_projection(self):
        """Test that a sliced proxy can be created for grids."""
        dataset = DAPHandler("http://localhost:8001/?SimpleGrid[0]", self.app1).dataset

        self.assertEqual(dataset.SimpleGrid.x.data.shape, (1,))
        self.assertEqual(dataset.SimpleGrid.x.data.slice, (slice(0, 1, 1),))
        self.assertEqual(dataset.SimpleGrid.y.data.shape, (2,))
        self.assertEqual(dataset.SimpleGrid.y.data.slice, (slice(0, 3, 1),))
        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (1, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice, (slice(0, 1, 1), slice(0, 3, 1))
        )

    def test_base_type_with_projection(self):
        """Test that a sliced proxy can be created for a base type."""
        dataset = DAPHandler("http://localhost:8001/?x[1:1:2]", self.app1).dataset

        self.assertEqual(dataset.x.data.shape, (2,))
        self.assertEqual(dataset.x.data.slice, (slice(1, 3, 1),))

    def test_grid_array_with_projection(self):
        """Test that a grid array can be properly pre sliced."""
        dataset = DAPHandler(
            "http://localhost:8001/?SimpleGrid.SimpleGrid[0]", self.app1
        ).dataset

        # object should be a structure, not a grid
        self.assertEqual(list(dataset.keys()), ["SimpleGrid"])
        self.assertNotIsInstance(dataset.SimpleGrid, GridType)
        self.assertIsInstance(dataset.SimpleGrid, StructureType)

        self.assertEqual(dataset.SimpleGrid.SimpleGrid.data.shape, (1, 3))
        self.assertEqual(
            dataset.SimpleGrid.SimpleGrid.data.slice, (slice(0, 1, 1), slice(0, 3, 1))
        )

    def test_grid_map_with_projection(self):
        """Test that a grid map can be properly pre sliced."""
        dataset = DAPHandler(
            "http://localhost:8001/?SimpleGrid.x[0]", self.app1
        ).dataset

        self.assertEqual(dataset.SimpleGrid.x.data.shape, (1,))
        self.assertEqual(dataset.SimpleGrid.x.data.slice, (slice(0, 1, 1),))

    def test_sequence(self):
        """Test that dataset has the correct data proxies for sequences."""
        dataset = DAPHandler("http://localhost:8001/", self.app2).dataset

        self.assertEqual(list(dataset.keys()), ["cast"])
        self.assertEqual(
            list(dataset.cast.keys()),
            [
                "id",
                "lon",
                "lat",
                "depth",
                "time",
                "temperature",
                "salinity",
                "pressure",
            ],
        )

        # check the sequence
        self.assertIsInstance(dataset.cast.data, SequenceProxy)
        self.assertEqual(dataset.cast.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.cast.data.id, "cast")
        self.assertEqual(dataset.cast.data.shape, ())
        self.assertEqual(dataset.cast.data.selection, [])
        self.assertEqual(dataset.cast.data.slice, (slice(None),))

        # check a child
        self.assertIsInstance(dataset.cast.lon.data, SequenceProxy)
        self.assertEqual(dataset.cast.lon.data.baseurl, "http://localhost:8001/")
        self.assertEqual(dataset.cast.lon.data.id, "cast.lon")
        self.assertEqual(dataset.cast.lon.data.shape, ())
        self.assertEqual(dataset.cast.lon.data.dtype, np.dtype(">i4"))
        self.assertEqual(dataset.cast.lon.data.selection, [])
        self.assertEqual(dataset.cast.lon.data.slice, (slice(None),))

    def test_sequence_with_projection(self):
        """Test projections applied to sequences."""
        dataset = DAPHandler("http://localhost:8001/?cast[1]", self.app2).dataset

        self.assertEqual(dataset.cast.data.slice, (slice(1, 2, 1),))
        self.assertEqual(
            [tuple(row) for row in dataset.cast.iterdata()],
            [("2", 200, 10, 500, 1, 15, 35, 100)],
        )

    def test_custom_timeout_BaseProxyDap2(self):
        dataset = DAPHandler("http://localhost:8001/", self.app1, timeout=300).dataset

        for var in walk(dataset, pydap.model.BaseType):
            assert var.data.timeout == 300

    def test_custom_timeout_SequenceProxy(self):
        dataset = DAPHandler("http://localhost:8001/", self.app2, timeout=300).dataset
        for var in walk(dataset, pydap.model.SequenceType):
            assert var.data.timeout == 300

    def test_custom_timeout_BaseProxyDap4(self):

        dataset = DAPHandler(
            "http://localhost:8001/", self.app4, timeout=300, protocol="dap4"
        ).dataset

        for var in walk(dataset, pydap.model.BaseType):
            assert var.data.timeout == 300


class TestBaseProxy(unittest.TestCase):
    """Test `BaseProxy` objects."""

    def setUp(self):
        """Create a WSGI app"""
        self.app = BaseHandler(SimpleArray)

        self.data = BaseProxyDap2(
            "http://localhost:8001/", "byte", np.dtype("B"), (5,), application=self.app
        )

    def test_repr(self):
        """Test the object representation."""
        self.assertEqual(
            repr(self.data),
            "BaseProxy('http://localhost:8001/', 'byte', "
            "dtype('uint8'), (5,), "
            "(slice(None, None, None),))",
        )

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], np.arange(5))

    def test_getitem_out_of_bound(self):
        """
        Test the ``__getitem__`` method with out of bounds
        slices.
        The same result as numpy is expected.
        """
        np.testing.assert_array_equal(self.data[:10], np.arange(5)[:10])

    def test_inexact_division(self):
        """Test data retrieval when step > 1 and division inexact."""
        np.testing.assert_array_equal(self.data[0:3:2], np.arange(5)[0:3:2])

    def test_len(self):
        """Test length method."""
        self.assertEqual(len(self.data), 5)

    def test_iteration(self):
        """Test iteration."""
        self.assertEqual(list(iter(self.data)), list(range(5)))

    def test_comparisons(self):
        """Test all the comparisons."""
        np.testing.assert_array_equal(self.data == 2, np.arange(5) == 2)
        np.testing.assert_array_equal(self.data != 2, np.arange(5) != 2)
        np.testing.assert_array_equal(self.data >= 2, np.arange(5) >= 2)
        np.testing.assert_array_equal(self.data <= 2, np.arange(5) <= 2)
        np.testing.assert_array_equal(self.data > 2, np.arange(5) > 2)
        np.testing.assert_array_equal(self.data < 2, np.arange(5) < 2)


class TestBaseProxyShort(unittest.TestCase):
    """Test `BaseProxy` objects with short dtype."""

    def setUp(self):
        """Create a WSGI app with array data"""
        self.app = BaseHandler(SimpleArray)

        self.data = BaseProxyDap2(
            "http://localhost:8001/", "short", np.dtype(">h"), (), application=self.app
        )

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], np.array(1))
        assert self.data[:].dtype.char == "h"


class TestBaseProxyString(unittest.TestCase):
    """Test a ``BaseProxy`` with string data."""

    def setUp(self):
        """Create a WSGI app with array data"""
        dataset = DatasetType("test")
        dataset["s"] = BaseType("s", np.array(["one", "two", "three"]))
        self.app = BaseHandler(dataset)

        self.data = BaseProxyDap2(
            "http://localhost:8001/", "s", np.dtype("|S5"), (3,), application=self.app
        )

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(
            self.data[:], np.array(["one", "two", "three"], dtype="S")
        )


class TestSequenceProxy(unittest.TestCase):
    """Test that a ``SequenceProxy`` behaves like a Numpy structured array."""

    def setUp(self):
        """Create a WSGI app"""
        self.app = BaseHandler(VerySimpleSequence)
        self.local = VerySimpleSequence.sequence.data

        dataset = DAPHandler("http://localhost:8001/", self.app).dataset
        self.remote = dataset.sequence.data

    def test_repr(self):
        """Test the object representation."""
        self.assertEqual(
            repr(self.remote),
            "SequenceProxy('http://localhost:8001/', "
            "<SequenceType with children 'byte', 'int', 'float'>, [], "
            "(slice(None, None, None),))",
        )

    def test_attributes(self):
        """Test attributes of the remote sequence."""
        self.assertEqual(self.remote.baseurl, "http://localhost:8001/")
        self.assertEqual(self.remote.id, "sequence")
        self.assertEqual(list(self.remote.template.keys()), ["byte", "int", "float"])
        self.assertEqual(self.remote.selection, [])
        self.assertEqual(self.remote.slice, (slice(None),))

    def test_getitem(self):
        """Test modifications to the proxy object."""
        child = self.remote["int"]
        self.assertEqual(child.id, "sequence.int")
        self.assertEqual(child.template.dtype, np.dtype(">i4"))
        self.assertEqual(child.template.shape, ())

        print(self.remote)
        child = self.remote[["float", "int"]]
        print(child)
        self.assertEqual(child.id, "sequence.float,sequence.int")
        self.assertEqual(list(child.template.keys()), ["float", "int"])

        child = self.remote[0]
        self.assertEqual(child.slice, (slice(0, 1, 1),))

        child = self.remote[0:2]
        self.assertEqual(child.slice, (slice(0, 2, 1),))

    def test_url(self):
        """Test URL generation."""
        self.assertEqual(self.remote.url, "http://localhost:8001/.dods?sequence")
        self.assertEqual(
            self.remote["int"].url, "http://localhost:8001/.dods?sequence.int"
        )
        self.assertEqual(
            self.remote[["float", "int"]].url,
            "http://localhost:8001/.dods?sequence.float,sequence.int",
        )

    def test_iter(self):
        """Test iteration."""
        self.assertEqual(
            [tuple(row) for row in self.local],
            [
                (0, 1, 10.0),
                (1, 2, 20.0),
                (2, 3, 30.0),
                (3, 4, 40.0),
                (4, 5, 50.0),
                (5, 6, 60.0),
                (6, 7, 70.0),
                (7, 8, 80.0),
            ],
        )

        self.assertEqual(
            [tuple(row) for row in self.remote],
            [
                (0, 1, 10.0),
                (1, 2, 20.0),
                (2, 3, 30.0),
                (3, 4, 40.0),
                (4, 5, 50.0),
                (5, 6, 60.0),
                (6, 7, 70.0),
                (7, 8, 80.0),
            ],
        )

    def test_iter_child(self):
        """Test iteration over a child."""
        child = self.local["byte"]
        self.assertEqual(list(child), [0, 1, 2, 3, 4, 5, 6, 7])

        child = self.remote["byte"]
        self.assertEqual(list(child), [0, 1, 2, 3, 4, 5, 6, 7])

    def test_iter_find_pattern(self):
        pattern = b"Data:\n"
        # Check in a simple iteration:
        string_iter = iter([b"blahblah ", b"Data:\n", b"blahblahblah"])
        last_chunk = find_pattern_in_string_iter(pattern, string_iter)
        assert last_chunk == b""
        assert next(string_iter) == b"blahblahblah"

        # Check in a more complex iteration:
        string_iter = iter([b"blahblah Da", b"ta:\nblahb", b"lahblah"])
        last_chunk = find_pattern_in_string_iter(pattern, string_iter)
        assert last_chunk == b"blahb"
        assert next(string_iter) == b"lahblah"

        # Check for a character by character iteration:
        string_iter = iter(
            [
                b"b",
                b"l",
                b"a",
                b"h",
                b"b",
                b"l",
                b"a",
                b"h",
                b" ",
                b"D",
                b"a",
                b"t",
                b"a",
                b":",
                b"\n",
                b"b",
                b"l",
                b"a",
            ]
        )
        last_chunk = find_pattern_in_string_iter(pattern, string_iter)
        assert last_chunk == b""
        assert next(string_iter) == b"b"

    def test_comparisons(self):
        """Test lazy comparisons on the object."""
        filtered = self.remote[self.remote["byte"] == 4]
        self.assertEqual(filtered.selection, ["sequence.byte=4"])

        filtered = self.remote[self.remote["byte"] != 4]
        self.assertEqual(filtered.selection, ["sequence.byte!=4"])

        filtered = self.remote[self.remote["byte"] >= 4]
        self.assertEqual(filtered.selection, ["sequence.byte>=4"])

        filtered = self.remote[self.remote["byte"] <= 4]
        self.assertEqual(filtered.selection, ["sequence.byte<=4"])


class TestSequenceWithString(unittest.TestCase):

    def setUp(self):
        """Create a WSGI app"""
        self.app = BaseHandler(SimpleSequence)
        self.local = SimpleSequence.cast.data

        dataset = DAPHandler("http://localhost:8001/", self.app).dataset
        self.remote = dataset.cast.data

    def test_iter(self):
        """Test iteration."""
        self.assertEqual(
            [tuple(row) for row in self.local],
            [("1", 100, -10, 0, -1, 21, 35, 0), ("2", 200, 10, 500, 1, 15, 35, 100)],
        )

        self.assertEqual(
            [tuple(row) for row in self.remote],
            [("1", 100, -10, 0, -1, 21, 35, 0), ("2", 200, 10, 500, 1, 15, 35, 100)],
        )

    def test_projection(self):
        """Test if we can select only a few variables."""
        filtered = self.local[["salinity", "depth"]]
        self.assertEqual([tuple(row) for row in filtered], [(35, 0), (35, 500)])

        filtered = self.remote[["salinity", "depth"]]
        self.assertEqual([tuple(row) for row in filtered], [(35, 0), (35, 500)])

    def test_iter_child(self):
        """Test iteration over a child."""
        child = self.local["salinity"]
        self.assertEqual(list(child), [35, 35])

        child = self.remote["salinity"]
        self.assertEqual(list(child), [35, 35])

    def test_filtering(self):
        """Test filtering of data."""
        filtered = self.local[self.local["lon"] > 100]
        self.assertEqual(
            [tuple(row) for row in filtered], [("2", 200, 10, 500, 1, 15, 35, 100)]
        )

        filtered = self.remote[self.remote["lon"] > 100]
        self.assertEqual(
            [tuple(row) for row in filtered], [("2", 200, 10, 500, 1, 15, 35, 100)]
        )

        # filtering works because we store comparisons as lazy objects
        self.assertIsInstance(self.remote["lon"] > 100, ConstraintExpression)
        self.assertEqual(filtered.selection, ["cast.lon>100"])


class TestStringBaseType(unittest.TestCase):
    """Regression test for string base type."""

    def setUp(self):
        """Create a WSGI app with array data"""
        dataset = DatasetType("test")
        data = np.array("This is a test", dtype="S")
        dataset["s"] = BaseType("s", data)
        self.app = BaseHandler(dataset)

        self.data = BaseProxyDap2(
            "http://localhost:8001/", "s", np.dtype("|S14"), (), application=self.app
        )

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], "This is a test")


class TestArrayStringBaseType(unittest.TestCase):
    """Regression test for an array of unicode base type."""

    def setUp(self):
        """Create a WSGI app with array data"""
        dataset = DatasetType("test")
        self.original_data = np.array([["This ", "is "], ["a ", "test"]], dtype="<U5")
        dataset["s"] = BaseType("s", self.original_data)
        self.app = BaseHandler(dataset)

        self.data = DAPHandler("http://localhost:8001/", self.app).dataset.s

    def test_getitem(self):
        """Test the ``__getitem__`` method."""
        np.testing.assert_array_equal(self.data[:], self.original_data)



"""Basic functions related to the DAP spec."""

import operator
from functools import reduce
from itertools import zip_longest
from sys import maxsize as MAXSIZE

import numpy as np
from requests.utils import quote as quote_
from requests.utils import unquote as unquote_

from . import __version__
from .exceptions import ConstraintExpressionError

__dap__ = "2.15"

# when installed in --editable mode, the `__version__` can be a very long str
# which causes `test_responses_html.py to fail (Content-Length' != '5864'
# in test_headers). This will work for now.

__version__ = str(__version__)[:5]  # it used to be 3.4.1, len=5.

START_OF_SEQUENCE = b"\x5a\x00\x00\x00"
END_OF_SEQUENCE = b"\xa5\x00\x00\x00"
STRING = "|S128"
DEFAULT_TIMEOUT = 120  # 120 seconds = 2 minutes

NUMPY_TO_DAP2_TYPEMAP = {
    "d": "Float64",
    "f": "Float32",
    "h": "Int16",
    "H": "UInt16",
    "i": "Int32",
    "l": "Int32",
    "q": "Int32",
    "I": "UInt32",
    "L": "UInt32",
    "Q": "UInt32",
    # DAP2 does not support signed bytes.
    # Its Byte type is unsigned and thus corresponds
    # to numpy's 'B'.
    # The consequence is that there is no natural way
    # in DAP2 to represent numpy's 'b' type.
    # Ideally, DAP2 would have a signed Byte type
    # and an unsigned UByte type and we would have the
    # following mapping: {'b': 'Byte', 'B': 'UByte'}
    # but this not how the protocol has been defined.
    # This means that numpy's 'b' must be mapped to Int16
    # and data must be upconverted in the DODS response.
    "b": "Int16",
    "B": "Byte",
    # There are no boolean types in DAP2. Upconvert to
    # Byte:
    "?": "Byte",
    "S": "String",
    # Map numpy's 'U' to String b/c
    # DAP2 does not explicitly support unicode.
    "U": "String",
}

# DAP2 demands big-endian 32 bytes signed integers
# www.opendap.org/pdf/dap_2_data_model.pdf
# Before pydap 3.2.2, length was
# big-endian 32 bytes UNSIGNED integers:
# DAP2_ARRAY_LENGTH_NUMPY_TYPE = '>I'
# Since pydap 3.2.2, the length type is accurate:
DAP2_ARRAY_LENGTH_NUMPY_TYPE = ">i"

DAP2_TO_NUMPY_RESPONSE_TYPEMAP = {
    "Float64": ">d",
    "Float32": ">f",
    # This is a weird aspect of the DAP2 specification.
    # For backward-compatibility, Int16 and UInt16 are
    # encoded as 32 bits integers in the response,
    # respectively:
    "Int16": ">i",
    "UInt16": ">I",
    "Int32": ">i",
    "UInt32": ">I",
    # DAP2 does not support signed bytes.
    # It's Byte type is unsigned and thus corresponds
    # to numpy 'B'.
    # The consequence is that there is no natural way
    # in DAP2 to represent numpy's 'b' type.
    # Ideally, DAP2 would have a signed Byte type
    # and a usigned UByte type and we would have the
    # following mapping: {'Byte': 'b', 'UByte': 'B'}
    # but this not how the protocol has been defined.
    # This means that DAP2 Byte is unsigned and must be
    # mapped to numpy's 'B' type, usigned byte.
    "Byte": "B",
    # Map String to numpy's string type 'S' b/c
    # DAP2 does not explicitly support unicode.
    "String": "S",
    "URL": "S",
    #
    # These two types are not DAP2 but it is useful
    # to include them for compatiblity with other
    # data sources:
    "Int": ">i",
    "UInt": ">I",
}

# Typemap from lower case DAP2 types to
# numpy dtype string with specified endiannes.
# Here, the endianness is very important:
LOWER_DAP2_TO_NUMPY_PARSER_TYPEMAP = {
    "float64": ">d",
    "float32": ">f",
    "int16": ">h",
    "uint16": ">H",
    "int32": ">i",
    "uint32": ">I",
    "byte": "B",
    "string": STRING,
    "url": STRING,
    "int": ">i",
    "uint": ">I",
}

# Typemap from lower case DAP4 types to
# numpy dtype string with specified endiannes.
# Here, the endianness is very important:
DAP4_TO_NUMPY_PARSER_TYPEMAP = {
    "Float16": ">f2",
    "Float32": ">f4",
    "Float64": ">f8",
    "Int8": ">i1",
    "UInt8": ">u1",
    "Int16": ">i2",
    "UInt16": ">u2",
    "Int32": ">i4",
    "UInt32": ">u4",
    "Int64": ">i8",
    "UInt64": ">u8",
    "Byte": "B",
    "String": STRING,
    "Url": STRING,
    "Char": ">u1",
}


def _quote(name):
    """Return quoted name according to the DAP specification.

    >>> _quote("White space")
    'White%20space'
    >>> _quote("Period.")
    'Period%2E'

    """
    safe = "%_!~*'-\"/"
    if "dap4" == name[:4]:
        # Dap4 protocol. Must not scape = sign there.
        prefix = name[:8]
        name = name[8:]
    else:
        prefix = ""
    name = quote_(name.encode("utf-8"), safe=safe).replace(".", "%2E")
    return prefix + name.replace("[", "%5B").replace("]", "%5D")


def unquote(name):
    """Return unquoted name according to the DAP specification.

    >>> unquote("White%20space")
    'White space'
    >>> unquote("Period%2E")
    'Period.'

    """
    name = name.replace("%2E", ".").replace("%5B", "[").replace("%5D", "]")
    return unquote_(name)


def encode(obj):
    """Return an object encoded to its DAP representation."""
    # fix for Python 3.5, where strings are being encoded as numbers
    if isinstance(obj, str) or isinstance(obj, np.ndarray) and obj.dtype.char in "SU":
        return '"{0}"'.format(obj)

    # fix for DeprecationWarning: Conversion of an array with ndim > 0 to a
    # scalar is deprecated
    if isinstance(obj, np.ndarray) and np.ndim(obj) > 0:
        arr_str = np.array2string(obj, formatter={"float_kind": lambda x: f"{x:.6f}"})
        return f'"[{arr_str[1:-1]}]"'
    else:
        try:
            return "%.6g" % obj
        except Exception:
            return '"{0}"'.format(obj)


def fix_slice(slice_, shape):
    """Return a normalized slice.

    This function returns a slice so that it has the same length of `shape`,
    and no negative indexes, if possible.

    This is based on this document:

        https://numpy.org/doc/stable/user/basics.indexing.html#slicing-and-striding

    """
    # convert `slice_` to a tuple
    if not isinstance(slice_, tuple):
        slice_ = (slice_,)

    # expand Ellipsis and make `slice_` at least as long as `shape`
    expand = len(shape) - len(slice_)
    out = []
    for s in slice_:
        if s is Ellipsis:
            out.extend((slice(None),) * (expand + 1))
            expand = 0
        else:
            out.append(s)
    slice_ = tuple(out) + (slice(None),) * expand

    out = []
    for s, N in zip(slice_, shape):
        if isinstance(s, int):
            if s < 0:
                s += N
            out.append(s)
        else:
            k = s.step or 1

            i = s.start
            if i is None:
                i = 0
            elif i < 0:
                i += N

            j = s.stop
            if j is None or j >= (N + i):
                j = N + i
            elif j < 0:
                j += N

            out.append(slice(i, j, k))

    return tuple(out)


def combine_slices(slice1, slice2):
    """Return two tuples of slices combined sequentially.

    These two should be equal:

        x[ combine_slices(s1, s2) ] == x[s1][s2]

    """
    out = []
    for exp1, exp2 in zip_longest(slice1, slice2, fillvalue=slice(None)):
        if isinstance(exp1, int):
            exp1 = slice(exp1, exp1 + 1)
        if isinstance(exp2, int):
            exp2 = slice(exp2, exp2 + 1)

        start = (exp1.start or 0) + (exp2.start or 0)
        step = (exp1.step or 1) * (exp2.step or 1)

        if exp1.stop is None and exp2.stop is None:
            stop = None
        elif exp1.stop is None:
            stop = (exp1.start or 0) + exp2.stop
        elif exp2.stop is None:
            stop = exp1.stop
        else:
            stop = min(exp1.stop, (exp1.start or 0) + exp2.stop)

        out.append(slice(start, stop, step))
    return tuple(out)


def hyperslab(slice_):
    """Return a DAP representation of a multidimensional slice."""
    if not isinstance(slice_, tuple):
        slice_ = [slice_]
    else:
        slice_ = list(slice_)

    while slice_ and slice_[-1] == slice(None):
        slice_.pop(-1)

    return "".join(
        "[%s:%s:%s]" % (s.start or 0, s.step or 1, (s.stop or MAXSIZE) - 1)
        for s in slice_
    )


def walk(var, type=object):
    """Yield all variables of a given type from a dataset.

    The iterator returns also the parent variable.

    """
    if isinstance(var, type):
        yield var
    for child in var.children():
        for subvar in walk(child, type):
            yield subvar


def tree(template, prefix=""):
    # Print the current node's name, add '.' at the root level
    if prefix == "":
        print(f".{unquote_(template.name)}")
    else:
        print(template.name)

    # Iterate over the children
    Nchild = len([child for child in template.children()])
    for i, child in enumerate(template.children()):
        # Determine the prefix for the current child
        if i == Nchild - 1:
            child_prefix = "└──"
            next_prefix = "   "
        else:
            child_prefix = "├──"
            next_prefix = "│  "

        # Print the current child
        print(prefix + child_prefix, end="")

        # Recursively call tree on the child with updated prefix
        tree(child, prefix + next_prefix)


def fix_shorthand(projection, dataset):
    """Fix shorthand notation in the projection.

    Some clients request variables by their name, not by the id. This is called
    the "shorthand notation", and it has to be fixed. This function will return
    a new projection with no shorthand calls.

    """
    out = []
    for var in projection:
        if len(var) == 1 and var[0][0] not in list(dataset.keys()):
            token, slice_ = var.pop(0)
            for child in walk(dataset):
                if token == unquote(child.name):
                    if var:
                        raise ConstraintExpressionError(
                            "Ambiguous shorthand notation request: %s" % token
                        )
                    var = [(parent, ()) for parent in child.id.split(".")[:-1]] + [
                        (token, slice_)
                    ]
        out.append(var)
    return out


def get_var(dataset, id_):
    """Given an id, return the corresponding variable from the dataset."""
    tokens = id_.split(".")
    return reduce(operator.getitem, [dataset] + tokens)


def decode_np_strings(numpy_var):
    """Given a fixed-width numpy string, decode it to a unicode type"""
    if isinstance(numpy_var, bytes) and hasattr(numpy_var, "tobytes"):
        return numpy_var.tobytes().decode("utf-8")
    else:
        return numpy_var


class StreamReader(object):
    """Class to allow reading a `urllib3.HTTPResponse`."""

    def __init__(self, stream):
        self.stream = stream
        self.buf = bytearray()

    def read(self, n):
        """Read and return `n` bytes."""
        while len(self.buf) < n:
            bytes_read = next(self.stream)
            self.buf.extend(bytes_read)

        out = bytes(self.buf[:n])
        self.buf = self.buf[n:]
        return out


class BytesReader(object):
    """Class to allow reading a `bytes` object."""

    def __init__(self, data):
        self.data = data

    def read(self, n):
        """Read and return `n` bytes."""
        out = self.data[:n]
        self.data = self.data[n:]
        return out

    def peek(self, n):
        return self.data[:n]



"""Tests for the pydap server."""

import multiprocessing
import os
import shutil
import tempfile
import unittest
from xml.etree import ElementTree as etree

from webob import Response
from webob.dec import wsgify
from webtest import AppError
from webtest import TestApp as App

from pydap.exceptions import ExtensionNotSupportedError
from pydap.wsgi.app import DapServer, PyDapApplication, StaticMiddleware, init


class TestDapServer(unittest.TestCase):
    """Tests for the pydap server."""

    def setUp(self):
        """Create an installation."""
        self.install = tempfile.mkdtemp(suffix="pydap")

        # create directory for data with two files
        data = os.path.join(self.install, "data")
        os.mkdir(data)
        os.mkdir(os.path.join(data, "subdir"))
        with open(os.path.join(data, "README.txt"), "w") as fp:
            fp.write("Hello, world!")
        with open(os.path.join(data, "data.foo"), "w") as fp:
            pass

        # create templates directory
        templates = os.path.join(self.install, "templates")
        init(templates)

        app = DapServer(data, templates)
        app.handlers = [DummyHandler]
        app = StaticMiddleware(app, os.path.join(templates, "static"))
        self.app = App(app)

    def tearDown(self):
        """Remove the installation."""
        shutil.rmtree(self.install)

    def test_app_root(self):
        """Test a regular request."""
        res = self.app.get("/")
        self.assertEqual(res.status, "200 OK")

    def test_app_hack(self):
        """Test a request to a resource that is outside the root dir."""
        with self.assertRaises(AppError):
            self.app.get("/../../../../../../../etc/passwd")

    def test_app_file(self):
        """Test that we can download a file."""
        res = self.app.get("/README.txt")
        self.assertEqual(res.text, "Hello, world!")

    def test_dap_request(self):
        """Test that DAP requests work."""
        res = self.app.get("/data.foo.dds")
        self.assertEqual(res.text, "Success!")

    def test_invalid_dap_request(self):
        """Test invalid DAP requests."""
        with self.assertRaises(ExtensionNotSupportedError):
            self.app.get("/README.txt.dds")

    def test_thredds_catalog_request(self):
        """Test that THREDDS Catalog requests work."""
        res = self.app.get("/catalog.xml")
        self.assertEqual(res.status, "200 OK")
        self.assertTrue(res.text.startswith("<?xml"))
        xml = etree.fromstring(res.text)
        self.assertEqual(
            xml.tag,
            "{http://www.unidata.ucar.edu/namespaces/thredds/InvCatalog/v1.0}catalog",
        )  # noqa

    def test_not_found(self):
        """Test 404 responses."""
        with self.assertRaises(AppError):
            self.app.get("/README.html")

    def test_asset(self):
        """Test that we can load a static asset."""
        res = self.app.get("/static/style.css")
        self.assertEqual(res.status, "200 OK")


class TestPyDapApplication(unittest.TestCase):
    """tests the configuration of Application"""

    def setUp(self):
        """Create an installation."""
        self.install = tempfile.mkdtemp(suffix="pydap")

        # create directory for data with two files
        data = os.path.join(self.install, "data")
        os.mkdir(data)
        os.mkdir(os.path.join(data, "subdir"))
        with open(os.path.join(data, "README.txt"), "w") as fp:
            fp.write("Hello, world!")
        with open(os.path.join(data, "data.foo"), "w") as fp:
            pass

        # create templates directory
        templates = os.path.join(self.install, "templates")
        init(templates)

        app = DapServer(data, templates)
        app.handlers = [DummyHandler]
        app = StaticMiddleware(app, os.path.join(templates, "static"))
        nworks = multiprocessing.cpu_count() // 2
        nthreads = multiprocessing.cpu_count() // 2
        self.app = PyDapApplication(
            app, host="127.0.1", port="8001", workers=nworks, threads=nthreads
        )

    def tearDown(self):
        """Remove the installation."""
        shutil.rmtree(self.install)

    def test_app_bind(self):
        """Test the correct config."""
        bind = self.app.cfg.settings["bind"].value[0]
        self.assertEqual(bind, "127.0.1:8001")

    def test_app_cfgworkers(self):
        workers = self.app.cfg.settings["workers"].value
        self.assertEqual(workers, multiprocessing.cpu_count() // 2)

    def test_app_cfgthreads(self):
        threads = self.app.cfg.settings["threads"].value
        self.assertEqual(threads, multiprocessing.cpu_count() // 2)


class TestPackageAssets(unittest.TestCase):
    """Test that we can load assets from the package."""

    def setUp(self):
        """Create an installation."""
        self.install = tempfile.mkdtemp(suffix="pydap")

        # create directory for data with two files
        data = os.path.join(self.install, "data")
        os.mkdir(data)
        with open(os.path.join(data, "README.txt"), "w") as fp:
            fp.write("Hello, world!")
        with open(os.path.join(data, "data.foo"), "w") as fp:
            pass

        app = DapServer(data)
        app = StaticMiddleware(app, ("pydap.wsgi", "templates/static"))
        self.app = App(app)

    def tearDown(self):
        """Remove the installation."""
        shutil.rmtree(self.install)

    def test_asset(self):
        """Test that we can load a static asset."""
        res = self.app.get("/static/style.css")
        self.assertEqual(res.status, "200 OK")

    def test_not_found(self):
        """Test 404 responses."""
        with self.assertRaises(AppError):
            self.app.get("/static/missing")


class DummyHandler(object):
    """A dummy handler for testing the server."""

    extensions = r"^.*\.foo$"

    def __init__(self, filepath):
        pass

    @wsgify
    def __call__(self, req):
        return Response("Success!")



"""A pydap handler for CSV files."""

import copy
import csv
import json
import os
import re
import time
from email.utils import formatdate
from stat import ST_MTIME

from ...exceptions import OpenFileError
from ...handlers.lib import BaseHandler, IterData
from ...model import BaseType, DatasetType, SequenceType
from ...parsers.das import add_attributes


class CSVHandler(BaseHandler):
    """This is a simple handler for CSV files.

    Here's a standard dataset for testing sequential data:

        >>> data = [
        ... (10, 15.2, 'Diamond_St'),
        ... (11, 13.1, 'Blacktail_Loop'),
        ... (12, 13.3, 'Platinum_St'),
        ... (13, 12.1, 'Kodiak_Trail')]

        >>> import csv
        >>> temp_file = getfixture('tmpdir').join('test.csv')
        >>> with open(str(temp_file), 'w') as f:
        ...     writer = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC)
        ...     row_length = writer.writerow(['index', 'temperature', 'site'])
        ...     row_length = [writer.writerow(row) for row in data]

    Iteraring over the sequence returns data:

        >>> seq = CSVHandler(str(temp_file)).dataset['sequence']

        >>> for line in seq:
        ...     print(line)
        (10.0, 15.2, 'Diamond_St')
        (11.0, 13.1, 'Blacktail_Loop')
        (12.0, 13.3, 'Platinum_St')
        (13.0, 12.1, 'Kodiak_Trail')

    The order of the variables can be changed:

        >>> for line in seq['temperature', 'site', 'index']:
        ...     print(line)
        (15.2, 'Diamond_St', 10.0)
        (13.1, 'Blacktail_Loop', 11.0)
        (13.3, 'Platinum_St', 12.0)
        (12.1, 'Kodiak_Trail', 13.0)

    We can iterate over children:

        >>> for line in seq['temperature']:
        ...     print(line)
        15.2
        13.1
        13.3
        12.1

    We can filter the data:

        >>> for line in seq[ seq.index > 10 ]:
        ...     print(line)
        (11.0, 13.1, 'Blacktail_Loop')
        (12.0, 13.3, 'Platinum_St')
        (13.0, 12.1, 'Kodiak_Trail')

        >>> for line in seq[ seq.index > 10 ]['site']:
        ...     print(line)
        Blacktail_Loop
        Platinum_St
        Kodiak_Trail

        >>> for line in seq[['site', 'temperature']][ seq.index > 10 ]:
        ...     print(line)
        ('Blacktail_Loop', 13.1)
        ('Platinum_St', 13.3)
        ('Kodiak_Trail', 12.1)

    Or slice it:

        >>> for line in seq[::2]:
        ...     print(line)
        (10.0, 15.2, 'Diamond_St')
        (12.0, 13.3, 'Platinum_St')

        >>> for line in seq[ seq.index > 10 ][::2]['site']:
        ...     print(line)
        Blacktail_Loop
        Kodiak_Trail

        >>> for line in seq[ seq.index > 10 ]['site'][::2]:
        ...     print(line)
        Blacktail_Loop
        Kodiak_Trail

    Finally, delete the data file:

        >>> temp_file.remove()
    """

    extensions = re.compile(r"^.*\.csv$", re.IGNORECASE)

    def __init__(self, filepath):
        BaseHandler.__init__(self)

        try:
            with open(filepath, "r") as fp:
                reader = csv.reader(fp, quoting=csv.QUOTE_NONNUMERIC)
                vars = next(reader)
        except Exception as exc:
            message = "Unable to open file {filepath}: {exc}".format(
                filepath=filepath, exc=exc
            )
            raise OpenFileError(message)

        self.additional_headers.append(
            (
                "Last-modified",
                (formatdate(time.mktime(time.localtime(os.stat(filepath)[ST_MTIME])))),
            )
        )

        # build dataset
        name = os.path.split(filepath)[1]
        self.dataset = DatasetType(name)

        # add sequence and children for each column
        seq = self.dataset["sequence"] = SequenceType("sequence")
        for var in vars:
            seq[var] = BaseType(var)

        # set the data
        seq.data = CSVData(filepath, copy.copy(seq))

        # add extra attributes
        metadata = "{0}.json".format(filepath)
        if os.path.exists(metadata):
            with open(metadata) as fp:
                attributes = json.load(fp)
            add_attributes(self.dataset, attributes)


class CSVData(IterData):
    """Emulate a Numpy structured array using CSV files.

    Here's a standard dataset for testing sequential data:

        >>> data = [
        ... (10, 15.2, 'Diamond_St'),
        ... (11, 13.1, 'Blacktail_Loop'),
        ... (12, 13.3, 'Platinum_St'),
        ... (13, 12.1, 'Kodiak_Trail')]

        >>> import csv
        >>> temp_file = getfixture('tmpdir').join('test.csv')
        >>> with open(str(temp_file), 'w') as f:
        ...     writer = csv.writer(f, quoting=csv.QUOTE_NONNUMERIC)
        ...     row_length = writer.writerow(['index', 'temperature', 'site'])
        ...     row_length = [writer.writerow(row) for row in data]

    Iteraring over the sequence returns data:

        >>> seq = SequenceType('example')
        >>> seq['index'] = BaseType('index')
        >>> seq['temperature'] = BaseType('temperature')
        >>> seq['site'] = BaseType('site')
        >>> seq.data = CSVData(str(temp_file), copy.copy(seq))

        >>> for line in seq:
        ...     print(line)
        (10.0, 15.2, 'Diamond_St')
        (11.0, 13.1, 'Blacktail_Loop')
        (12.0, 13.3, 'Platinum_St')
        (13.0, 12.1, 'Kodiak_Trail')

    The order of the variables can be changed:

        >>> for line in seq['temperature', 'site', 'index']:
        ...     print(line)
        (15.2, 'Diamond_St', 10.0)
        (13.1, 'Blacktail_Loop', 11.0)
        (13.3, 'Platinum_St', 12.0)
        (12.1, 'Kodiak_Trail', 13.0)

    We can iterate over children:

        >>> for line in seq['temperature']:
        ...     print(line)
        15.2
        13.1
        13.3
        12.1

    We can filter the data:

        >>> for line in seq[ seq.index > 10 ]:
        ...     print(line)
        (11.0, 13.1, 'Blacktail_Loop')
        (12.0, 13.3, 'Platinum_St')
        (13.0, 12.1, 'Kodiak_Trail')

        >>> for line in seq[ seq.index > 10 ]['site']:
        ...     print(line)
        Blacktail_Loop
        Platinum_St
        Kodiak_Trail

        >>> for line in seq['site', 'temperature'][ seq.index > 10 ]:
        ...     print(line)
        ('Blacktail_Loop', 13.1)
        ('Platinum_St', 13.3)
        ('Kodiak_Trail', 12.1)

    Or slice it:

        >>> for line in seq[::2]:
        ...     print(line)
        (10.0, 15.2, 'Diamond_St')
        (12.0, 13.3, 'Platinum_St')

        >>> for line in seq[ seq.index > 10 ][::2]['site']:
        ...     print(line)
        Blacktail_Loop
        Kodiak_Trail

        >>> for line in seq[ seq.index > 10 ]['site'][::2]:
        ...     print(line)
        Blacktail_Loop
        Kodiak_Trail

    Finally, delete the data file:

        >>> temp_file.remove()

    """

    def __init__(
        self, filepath, template, ifilter=None, imap=None, islice=None, level=0
    ):
        self.filepath = filepath
        self.template = template
        self.level = level

        self.ifilter = ifilter or []
        self.imap = imap or []
        self.islice = islice or []

    @property
    def stream(self):
        """Generator that yield lines of the file."""
        try:
            with open(self.filepath, "r") as fp:
                reader = csv.reader(fp, quoting=csv.QUOTE_NONNUMERIC)
                next(reader)  # consume var names
                for row in reader:
                    yield row
        except Exception as exc:
            message = "Unable to open file {filepath}: {exc}".format(
                filepath=self.filepath, exc=exc
            )
            raise OpenFileError(message)

    def __copy__(self):
        """Return a lightweight copy."""
        return self.__class__(
            self.filepath,
            copy.copy(self.template),
            self.ifilter[:],
            self.imap[:],
            self.islice[:],
            self.level,
        )



"""The DAS response.

The DAS response describes the attributes associated with a dataset and its
variables. Together with the DDS the DAS response completely describes the
metadata of a dataset, allowing it to be introspected and data to be
downloaded.

"""

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from collections.abc import Iterable

import numpy as np

from ..lib import NUMPY_TO_DAP2_TYPEMAP, __version__, _quote, encode
from ..model import BaseType, DatasetType, GridType, SequenceType, StructureType
from .lib import BaseResponse

INDENT = " " * 4


class DASResponse(BaseResponse):
    """The DAS response."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_das"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        for line in das(self.dataset):
            try:
                yield line.encode("ascii")
            except UnicodeDecodeError:
                yield line.encode("UTF-8")


@singledispatch
def das(var, level=0):
    """Single dispatcher that generates the DAS response."""
    raise StopIteration


@das.register(DatasetType)
def _datasettype(var, level=0):
    yield "{indent}Attributes {{\n".format(indent=level * INDENT)

    for attr in sorted(var.attributes.keys()):
        values = var.attributes[attr]
        for line in build_attributes(attr, values, level + 1):
            yield line

    for child in var.children():
        for line in das(child, level=level + 1):
            yield line
    yield "{indent}}}\n".format(indent=level * INDENT)


@das.register(StructureType)
@das.register(SequenceType)
def _structuretype(var, level=0):
    yield "{indent}{name} {{\n".format(indent=level * INDENT, name=var.name)

    for attr in sorted(var.attributes.keys()):
        values = var.attributes[attr]
        for line in build_attributes(attr, values, level + 1):
            yield line

    for child in var.children():
        for line in das(child, level=level + 1):
            yield line
    yield "{indent}}}\n".format(indent=level * INDENT)


@das.register(BaseType)
@das.register(GridType)
def _basetypegridtype(var, level=0):
    yield "{indent}{name} {{\n".format(indent=level * INDENT, name=var.name)

    for attr in sorted(var.attributes.keys()):
        values = var.attributes[attr]
        if np.asarray(values).size > 0:
            for line in build_attributes(attr, values, level + 1):
                yield line
    yield "{indent}}}\n".format(indent=level * INDENT)


def build_attributes(attr, values, level=0):
    """Recursive function to build the DAS."""
    # check for metadata
    if isinstance(values, dict):
        yield "{indent}{attr} {{\n".format(indent=(level) * INDENT, attr=attr)
        for k, v in values.items():
            for line in build_attributes(k, v, level + 1):
                yield line
        yield "{indent}}}\n".format(indent=(level) * INDENT)
    else:
        # get type
        type = get_type(values)

        # encode values
        if (
            isinstance(values, str)
            or not isinstance(values, Iterable)
            or getattr(values, "shape", None) == ()
        ):
            values = [encode(values)]
        else:
            values = map(encode, values)

        yield "{indent}{type} {attr} {values};\n".format(
            indent=(level) * INDENT,
            type=type,
            attr=_quote(attr),
            values=", ".join(values),
        )


def get_type(values):
    """Extract the type of a variable.

    This function tries to determine the DAP type of a Python variable using
    several methods. Returns the DAP type as a string.

    """
    if hasattr(values, "dtype"):
        return NUMPY_TO_DAP2_TYPEMAP[values.dtype.char]
    elif isinstance(values, str) or not isinstance(values, Iterable):
        return type_convert(values)
    else:
        # if there are several values, they may have different types, so we
        # need to convert all of them and use a precedence table
        types = [type_convert(val) for val in values]
        precedence = ["String", "Float64", "Int32"]
        types.sort(key=precedence.index)
        return types[0]


def type_convert(obj):
    """Map Python objects to the corresponding Opendap types.

    Returns the DAP representation of the type as a string.

    """
    if isinstance(obj, float):
        return "Float64"
    elif isinstance(obj, int):
        return "Int32"
    else:
        return "String"



"""Test DAP4 parsing functions."""

import os
import re
import unittest
from xml.etree import ElementTree as ET

import numpy as np

from ..lib import walk
from ..model import BaseType
from ..parsers.dmr import dmr_to_dataset, get_groups


def load_dmr_file(file_path):
    abs_path = os.path.join(os.path.dirname(__file__), file_path)
    with open(abs_path, "r") as dmr_file:
        dmr = dmr_file.read()
    dataset = dmr_to_dataset(dmr)
    return dataset


class TestDMRParser(unittest.TestCase):
    """Test parsing a DMR."""

    def test_single_scalar(self):
        """Test a single scalar case."""
        single_scalar_dmr = """<Dataset name="foo"><Int32 name="x"/></Dataset>"""
        dataset = dmr_to_dataset(single_scalar_dmr)
        self.assertEqual(dataset["x"].dtype, ">i4")

    def test_missing_value(self):
        """Test cases with missing value."""
        NAN_dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="missing_value" type="Float32">\n
                            <Value>NaN</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        INF_dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="missing_value" type="Float32">\n
                            <Value>Inf</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""

        dataset_nan = dmr_to_dataset(NAN_dmr)
        dataset_inf = dmr_to_dataset(INF_dmr)

        X_nan = dataset_nan["x"]
        X_inf = dataset_inf["x"]
        self.assertEqual(np.isnan(X_nan.attributes["missing_value"]), True)
        self.assertEqual(np.isinf(X_inf.attributes["missing_value"]), True)

    def test_attribute_warning(self):
        byte_dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="missing_value" type="Byte">\n
                            <Value>x00</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        with self.assertRaises(Warning):
            dmr_to_dataset(byte_dmr)

    def test_coads_climatology2(self):
        dataset = load_dmr_file("data/dmrs/coads_climatology.nc.dmr")
        self.assertEqual(
            dataset["SST"].attributes["long_name"], "SEA SURFACE TEMPERATURE"
        )
        self.assertEqual(dataset["SST"].attributes["missing_value"], -9.99999979e33)
        self.assertEqual(dataset["AIRT"].shape, (12, 90, 180))
        self.assertEqual(dataset["VWND"].dtype.str, ">f4")

    def test_atl03(self):
        # Contains groups
        dataset = load_dmr_file(
            "data/dmrs/ATL03_20181228015957_13810110_003_01.2var.h5.dmrpp.dmr"
        )
        self.assertEqual(
            dataset["/gt1r/bckgrd_atlas/bckgrd_int_height"].attributes["contentType"],
            "modelResult",
        )

    def test_jpl(self):
        # Contains groups
        dataset = load_dmr_file(
            "data/dmrs/20220102090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1.dmr"
        )
        self.assertEqual(
            dataset["sea_ice_fraction"].dimensions, ["/time", "/lat", "/lon"]
        )
        self.assertEqual(dataset.attributes["Conventions"], "CF-1.7")

    def test_mod05(self):
        dataset = load_dmr_file(
            "data/dmrs/MOD05_L2.A2019336.2315.061.2019337071952.hdf.dmr"
        )
        self.assertEqual(
            dataset["Water_Vapor_Infrared"].dimensions,
            [
                "/Cell_Along_Swath_5km",
                "/Cell_Across_Swath_5km",
            ],
        )

    def test_SWOT(self):
        dataset = load_dmr_file("data/dmrs/SWOT_GPR.dmr")
        # pick a single variable Maps
        maps = ("/data_01/longitude", "/data_01/latitude")
        self.assertEqual(dataset["data_01/ku/swh_ocean"].Maps, maps)

    def tests_global_dimensions(self):
        dataset = load_dmr_file("data/dmrs/SimpleGroup.dmr")
        # pick a single variable Maps
        names = [key for key in dataset.dimensions.keys()]
        sizes = [v[1] for v in dataset.dimensions.items()]
        self.assertEqual(names, ["time", "nv"])
        self.assertEqual(sizes, [1, 2])

    def tests_named_dimension(self):
        dataset = load_dmr_file("data/dmrs/SimpleGroup.dmr")
        # get only names of dimensions
        names = [key for key in dataset.dimensions.keys()]
        # get all variables/arrays
        variables = []
        for var in walk(dataset, BaseType):
            variables.append(var.name)
        # assert nv is a Global dimension
        self.assertIn("nv", names)
        # assert nv is NOT a variable/array
        self.assertNotIn("nv", variables)

    def tests_FlatGroups(self):
        dataset = load_dmr_file("data/dmrs/SimpleGroupFlat.dmr")
        # pick a single variable Maps
        Groups = dataset.groups()
        Variables = [item for (item, _) in dataset.variables().items()]
        self.assertEqual(Groups, {})
        self.assertEqual(
            Variables,
            [
                "SimpleGroup/Temperature",
                "SimpleGroup/Salinity",
                "SimpleGroup/Y",
                "SimpleGroup/X",
                "time",
                "time_bnds",
            ],
        )

    def test_get_groups(self):
        dmr_file = "data/dmrs/SimpleGroup.dmr"
        abs_path = os.path.join(os.path.dirname(__file__), dmr_file)
        with open(abs_path, "r") as dmr_file:
            dmr = dmr_file.read()
        _dmr = re.sub(' xmlns="[^"]+"', "", dmr, count=1)
        node = ET.fromstring(_dmr)
        groups = get_groups(node)
        assert isinstance(groups, dict)
        assert list(groups.keys()) == ["/SimpleGroup"]
        entry = groups["/SimpleGroup"]
        assert entry["attributes"] == {"Description": "Test group with numerical data"}
        assert entry["dimensions"] == {"Y": 4, "X": 4}
        assert entry["path"] == "/"
        assert entry["Maps"] == ()


class TestAttrsTypesDMRParser(unittest.TestCase):
    """Test parsing a DMR with all types"""

    def test_int8(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Int8">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_uint8(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="UInt8">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_Char(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Char">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_int16(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Int16">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_uint16(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="UInt16">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_int32(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Int32">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_uint32(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="UInt32">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_int64(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Int64">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_uint64(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="UInt64">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], int)

    def test_float32(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Float32">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], float)

    def test_float64(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Float64">\n
                            <Value>1</Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], float)

    def test_floatNone(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Float32">\n
                            <Value></Value>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert ds["x"].attributes["attr"] is None

    def test_TDSfloat64(self):
        dmr = """<Dataset name="foo">\n    <Int32 name="x">\n
                <Attribute name="attr" type="Float64">\n
                            <Value value="0.0"/>\n        </Attribute>\n
                                </Int32>\n</Dataset>"""
        ds = dmr_to_dataset(dmr)
        assert isinstance(ds["x"].attributes["attr"], float)



"""DAP exceptions.

These exceptions are mostly used by the server. When an exception is captured,
proper error message is displayed (according to the DAP 2.0 spec), with
information about the exception.

"""


class DapError(Exception):
    """Base DAP exception."""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ClientError(DapError):
    """Generic error with the client."""

    pass


class ServerError(DapError):
    """Generic error with the server."""

    pass


class ConstraintExpressionError(ServerError):
    """Exception raised when an invalid constraint expression is given."""

    pass


class HandlerError(DapError):
    """Generic error with a handler."""

    pass


class ExtensionNotSupportedError(HandlerError):
    """Exception raised when opening a file not supported by any handlers."""

    pass


class OpenFileError(HandlerError):
    """Exception raised when unable to open a file."""

    pass



"""Tests for the server-side function WSGI middleware."""

import sys
import unittest

import numpy as np
from webtest import TestApp as App

from pydap.exceptions import ServerError
from pydap.handlers.lib import BaseHandler
from pydap.model import BaseType, SequenceType
from pydap.tests.datasets import SimpleGrid, SimpleSequence
from pydap.wsgi.ssf import ServerSideFunctions


class TestMiddleware(unittest.TestCase):
    """Tests for the server-side function middleware."""

    def test_das(self):
        """Test that DAS requests are ignored."""
        # create a simple app
        app = App(ServerSideFunctions(BaseHandler(SimpleSequence)))

        # test a DAS response with a non-existing function
        app.get("/.das?non_existing_function(sequence)")

        # now test a DDS response
        with self.assertRaises(KeyError):
            app.get("/.dds?non_existing_function(sequence)")

    def test_no_parsed_response(self):
        """Test that non-parsed responses work or raise error.

        pydap returns WSGI responses that contain the "parsed" dataset, so it
        can be manipulated by middleware.

        """
        app = App(ServerSideFunctions(Accumulator(BaseHandler(SimpleGrid))))

        # a normal request should work, even though server-side functions are
        # not working in the WSGI pipeline
        app.get("/.dds")

        # this will fail, since it's impossible to build the response
        with self.assertRaises(ServerError):
            app.get("/.dds?mean(x)")

    def test_selection(self):
        """Test a selection server-side function."""
        app = App(ServerSideFunctions(BaseHandler(SimpleSequence)))
        res = app.get(
            "/.asc?density(cast.salinity,cast.temperature,cast.pressure)>1025"
        )
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        String id;
        Int32 lon;
        Int32 lat;
        Int32 depth;
        Int32 time;
        Int32 temperature;
        Int32 salinity;
        Int32 pressure;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.id, cast.lon, cast.lat, cast.depth, cast.time, cast.temperature, cast.salinity, cast.pressure
"2", 200, 10, 500, 1, 15, 35, 100

""",
        )

    def test_operators(self):
        """Test different operators on selection using a dummy function."""
        app = App(ServerSideFunctions(BaseHandler(SimpleSequence), double=double))
        res = app.get("/.asc?cast.lat&double(cast.lat)>10")
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat
10

""",
        )

        res = app.get("/.asc?cast.lat&double(cast.lat)>=20")
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat
10

""",
        )

        res = app.get("/.asc?cast.lat&double(cast.lat)<10")
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat
-10

""",
        )

        res = app.get("/.asc?cast.lat&double(cast.lat)<=-20")
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat
-10

""",
        )

        res = app.get("/.asc?cast.lat&double(cast.lat)=20")
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat
10

""",
        )

        res = app.get("/.asc?cast.lat&double(cast.lat)!=20")
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat
-10

""",
        )

    def test_selection_no_comparison(self):
        """Test function calls in the selection without comparison.

        Some functions, like ``bounds``, have no comparison. In this case, the
        selection is implicitely applied by comparing the result to 1.

        """
        app = App(ServerSideFunctions(BaseHandler(SimpleSequence)))
        res = app.get(
            "/.asc?cast.lat,cast.lon&"
            "bounds(0,360,-90,0,0,500,00Z01JAN1900,00Z01JAN2000)"
        )
        self.assertEqual(
            res.text,
            """Dataset {
    Sequence {
        Int32 lat;
        Int32 lon;
    } cast;
} SimpleSequence;
---------------------------------------------
cast.lat, cast.lon
-10, 100

""",
        )

    def test_projection(self):
        """Test a simple function call on a projection."""
        app = App(ServerSideFunctions(BaseHandler(SimpleGrid)))
        res = app.get("/.asc?mean(x)")
        self.assertEqual(
            res.text,
            """Dataset {
    Float64 x;
} SimpleGrid;
---------------------------------------------
x
1
""",
        )

    def test_projection_clash(self):
        """Test a function call creating a variable with a conflicting name."""
        app = App(ServerSideFunctions(BaseHandler(SimpleGrid)))
        res = app.get("/.asc?mean(x),x")
        self.assertEqual(
            res.text,
            """Dataset {
    Int32 x[x = 3];
} SimpleGrid;
---------------------------------------------
x
[0] 0
[1] 1
[2] 2

""",
        )

    def test_nested_projection(self):
        """Test a nested function call."""
        app = App(ServerSideFunctions(BaseHandler(SimpleGrid)))
        res = app.get("/.asc?mean(mean(SimpleGrid.SimpleGrid,0),0)")
        self.assertEqual(
            res.text,
            """Dataset {
    Float64 SimpleGrid;
} SimpleGrid;
---------------------------------------------
SimpleGrid
2.5
""",
        )


class Accumulator(object):
    """A WSGI middleware that breaks streaming."""

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        return list(self.app(environ, start_response))


def double(dataset, var):
    """A dummy function that doubles a value.

    The value must be in a sequence. Return a new sequence with the value
    doubled.

    """
    # sequence is the first variable
    sequence = next(dataset.children())

    # get a single variable and double its value
    selection = sequence[var.name]
    rows = [(value * 2,) for value in selection]

    # create output sequence
    out = SequenceType("result")
    out["double"] = BaseType("double")
    out.data = np.rec.fromrecords(rows, names=["double"])

    return out



import ssl
from contextlib import closing

import requests
from requests.exceptions import InvalidSchema, MissingSchema, Timeout
from requests.utils import urlparse, urlunparse
from webob.exc import HTTPError
from webob.request import Request

from .lib import DEFAULT_TIMEOUT, _quote


def GET(url, application=None, session=None, timeout=DEFAULT_TIMEOUT, verify=True):
    """Open a remote URL returning a webob.response.Response object

    Optional parameters:
    session: a requests.Session() object (potentially) containing
             authentication cookies.

    Optionally open a URL to a local WSGI application
    """
    if application:
        _, _, path, _, query, fragment = urlparse(url)
        url = urlunparse(("", "", path, "", _quote(query), fragment))

    response = follow_redirect(
        url, application=application, session=session, timeout=timeout, verify=verify
    )
    # Decode request response (i.e. gzip)
    response.decode_content()
    return response


def raise_for_status(response):
    # Raise error if status is above 300:
    if response.status_code >= 400:
        raise HTTPError(
            detail=response.status + "\n" + response.text,
            headers=response.headers,
            comment=response.body,
        )
    elif response.status_code >= 300:
        try:
            text = response.text
        except AttributeError:
            # With this status_code, response.text could
            # be ill-defined. If the redirect does not set
            # an encoding (i.e. response.charset is None).
            # Set the text to empty string:
            text = ""
        raise HTTPError(
            detail=(
                response.status
                + "\n"
                + text
                + "\n"
                + "This is redirect error. These should not usually raise "
                + "an error in pydap beacuse redirects are handled "
                + "implicitly. If it failed it is likely due to a "
                + "circular redirect."
            ),
            headers=response.headers,
            comment=response.body,
        )


def follow_redirect(
    url, application=None, session=None, timeout=DEFAULT_TIMEOUT, verify=True
):
    """
    This function essentially performs the following command:
    >>> Request.blank(url).get_response(application)  # doctest: +SKIP

    It however makes sure that the request possesses the same cookies and
    headers as the passed session.
    """

    req = create_request(url, session=session, timeout=timeout, verify=verify)
    return get_response(req, application, verify=verify)


def get_response(req, application, verify=True):
    """
    If verify=False, use the ssl library to temporarily disable
    ssl verification.
    """
    if verify:
        resp = req.get_response(application)
    else:
        # Here, we use monkeypatching. Webob does not provide a way
        # to bypass SSL verification.
        # This approach is never ideal but it appears to be the only option
        # here.
        # This only works in python 2.7 and >=3.5. Python 3.4
        # does not require it because by default contexts are not
        # verified.
        try:
            _create_default_https_ctx = ssl._create_default_https_context
            _create_unverified_ctx = ssl._create_unverified_context
            ssl._create_default_https_context = _create_unverified_ctx
        except AttributeError:
            _create_default_https_ctx = None

        try:
            resp = req.get_response(application)
        finally:
            if _create_default_https_ctx is not None:
                # Restore verified context
                ssl._create_default_https_context = _create_default_https_ctx
    return resp


def create_request(url, session=None, timeout=DEFAULT_TIMEOUT, verify=True):
    if session is not None:
        # If session is set and cookies were loaded using pydap.cas.get_cookies
        # using the check_url option, then we can legitimately expect that
        # the connection will go through seamlessly. However, there might be
        # redirects that might want to modify the cookies. Webob is not
        # really up to the task here. The approach used here is to
        # piggy back on the requests library and use it to fetch the
        # head of the requested url. Requests will follow redirects and
        # adjust the cookies as needed. We can then use the final url and
        # the final cookies to set up a webob Request object that will
        # be guaranteed to have all the needed credentials:
        return create_request_from_session(url, session, timeout=timeout, verify=verify)
    else:
        # If a session object was not passed, we simply pass a new
        # requests.Session() object. The requests library allows the
        # handling of redirects that are not naturally handled by Webob.
        return create_request_from_session(
            url, requests.Session(), timeout=timeout, verify=verify
        )


def create_request_from_session(url, session, timeout=DEFAULT_TIMEOUT, verify=True):
    try:
        # Use session to follow redirects:
        with closing(
            session.head(url, allow_redirects=True, timeout=timeout, verify=verify)
        ) as head:
            req = Request.blank(head.url)
            req.environ["webob.client.timeout"] = timeout

            # Get cookies from head:
            cookies_dict = head.cookies.get_dict()

            # Set request cookies to the head cookies:
            req.headers["Cookie"] = ",".join(
                name + "=" + cookies_dict[name] for name in cookies_dict
            )
            # Set the headers to the session headers:
            for item in head.request.headers:
                req.headers[item] = head.request.headers[item]
            return req
    except (MissingSchema, InvalidSchema):
        # Missing schema can occur in tests when the url
        # is not pointing to any resource. Simply pass.
        req = Request.blank(url)
        req.environ["webob.client.timeout"] = timeout
        return req
    except Timeout:
        raise HTTPError("Timeout")



import copy
import warnings

import requests
from bs4 import BeautifulSoup
from requests.packages.urllib3.exceptions import (
    InsecurePlatformWarning,
    InsecureRequestWarning,
)
from requests.utils import urlparse, urlunparse

from ..lib import __version__

ssl_verify_categories = [InsecureRequestWarning, InsecurePlatformWarning]


def setup_session(
    uri,
    username=None,
    password=None,
    check_url=None,
    session=None,
    verify=True,
    username_field="username",
    password_field="password",
):
    """
    A general function to set-up requests session with cookies
    using beautifulsoup and by calling the right url.
    """

    if session is None:
        # Connections must be closed since some CAS
        # will cough when connections are kept alive:
        headers = [
            ("User-agent", "pydap/{}".format(__version__)),
            ("Connection", "close"),
        ]
        session = requests.Session()
        session.headers.update(headers)

    if uri is None:
        return session

    if not verify:
        verify_flag = session.verify
        session.verify = False

    if isinstance(uri, str):
        url = uri
    else:
        url = uri(check_url)

    if password is None or password == "":
        warnings.warn(
            "password was not set. "
            "this was likely unintentional "
            "but will result is much fewer datasets."
        )
        if not verify:
            session.verify = verify_flag
        return session

    # Allow for several subsequent security layers:
    full_url = copy.copy(url)
    if isinstance(full_url, list):
        url = full_url[0]

    with warnings.catch_warnings():
        if not verify:
            # Catch warnings. It is assumed that the
            # user that explicitly uses verify=False
            # is either fully aware of the risks
            # or cannot avoid the risks because of
            # an improperly configured server.
            # This error will usually occur with
            # ESGF authentication.
            for category in ssl_verify_categories:
                warnings.filterwarnings("ignore", category=category)

        response = soup_login(
            session,
            url,
            username,
            password,
            username_field=username_field,
            password_field=password_field,
        )

        # If there are further security levels.
        # At the moment only used for CEDA OPENID:
        if isinstance(full_url, list) and len(full_url) > 1:
            for url in full_url[1:]:
                response = soup_login(
                    session,
                    response.url,
                    username,
                    password,
                    username_field=None,
                    password_field=None,
                )
        response.close()

        if check_url:
            if username is not None and password is not None:
                res = session.get(check_url, auth=(username, password))
                if res.status_code == 401:
                    res = session.get(res.url, auth=(username, password))
                res.close()
            raise_if_form_exists(check_url, session)

    if not verify:
        session.verify = verify_flag
    return session


def raise_if_form_exists(url, session):
    """
    This function raises a UserWarning if the link has forms
    """

    user_warning = (
        "Navigate to {0}, ".format(url) + "login and follow instructions. "
        "It is likely that you have to perform some one-time "
        "registration steps before acessing this data."
    )

    resp = session.get(url)
    soup = BeautifulSoup(resp.content, "lxml")
    if len(soup.select("form")) > 0:
        raise UserWarning(user_warning)


def soup_login(
    session,
    url,
    username,
    password,
    username_field="username",
    password_field="password",
):
    resp = session.get(url)

    soup = BeautifulSoup(resp.content, "lxml")
    login_form = soup.select("form")[0]

    def get_to_url(current_url, to_url):
        spl_current = urlparse(current_url)
        spl_to = urlparse(to_url)
        comb = []
        for val1, val2 in zip(spl_to, spl_current):
            if val1 == "":
                comb.append(val2)
            else:
                comb.append(val1)
        return urlunparse(comb)

    to_url = get_to_url(resp.url, login_form.get("action"))

    session.headers["Referer"] = resp.url

    payload = {}
    if username_field is not None:
        if len(login_form.findAll("input", {"name": username_field})) > 0:
            payload.update({username_field: username})

    if password_field is not None:
        if len(login_form.findAll("input", {"name": password_field})) > 0:
            payload.update({password_field: password})
        else:
            # If there is no password_field, it might be because
            # something should be handled in the browser
            # for the first attempt. This is common when using
            # pydap with the ESGF for the first time.
            raise Exception(
                "Navigate to {0}. "
                "If you are unable to "
                "login, you must either "
                "wait or use authentication "
                "from another service.".format(url)
            )

    # Replicate all other fields:
    for input in login_form.findAll("input"):
        if input.get("name") not in payload and input.get("name") is not None:
            payload.update({input.get("name"): input.get("value")})

    # Remove other submit fields:
    submit_type = "submit"
    submit_names = [
        input.get("name")
        for input in login_form.findAll("input", {"type": submit_type})
    ]
    for input in login_form.findAll("input", {"type": submit_type}):
        if "submit" in submit_names and input.get("name").lower() != "submit":
            payload.pop(input.get("name"), None)

    return session.post(to_url, data=payload)



"""This is the pydap data model, an implementation of the Data Access Protocol
data model written in Python.

The model is composed of a base object which represents data, the `BaseType`,
and by objects which can hold other objects, all derived from `StructureType`.
Here's a simple example of a `BaseType` variable::

    >>> import numpy as np
    >>> foo = BaseType('foo', np.arange(4, dtype='i'))
    >>> bar = BaseType('bar', np.arange(4, dtype='i'))
    >>> foobar = BaseType('foobar', np.arange(4, dtype='i'))
    >>> foo[-2:]
    <BaseType with data array([2, 3], dtype=int32)>
    >>> foo[-2:].data
    array([2, 3], dtype=int32)
    >>> foo.data[-2:]
    array([2, 3], dtype=int32)
    >>> foo.dtype
    dtype('int32')
    >>> foo.shape
    (4,)
    >>> for record in foo.iterdata():
    ...     print(record)
    0
    1
    2
    3

It is also possible to iterate directly over a `BaseType`::
    >>> for record in foo:
    ...     print(record)
    0
    1
    2
    3

This is however discouraged because this approach will soon be deprecated
for the `SequenceType` where only the ``.iterdata()`` will continue to be
supported.

The `BaseType` is simply a thin wrapper over Numpy arrays, implementing the
`dtype` and `shape` attributes, and the sequence and iterable protocols. Why
not use Numpy arrays directly then? First, `BaseType` can have additional
metadata added to them; this include names for its dimensions and also
arbitrary attributes::

    >>> foo.attributes
    {}
    >>> foo.attributes['units'] = 'm/s'
    >>> foo.units
    'm/s'

    >>> foo.dimensions
    ()
    >>> foo.dimensions = ('time',)

Second, `BaseType` can hold data objects other than Numpy arrays. There are
more complex data objects, like `pydap.handlers.dap.BaseProxy`, which acts as a
transparent proxy to a remote dataset, exposing it through the same interface.

Now that we have some data, we can organize it using containers::

    >>> dataset = DatasetType('baz')
    >>> dataset['s'] = StructureType('s')
    >>> dataset['s']['foo'] = foo
    >>> dataset['s']['bar'] = bar
    >>> dataset['s']['foobar'] = foobar

`StructureType` and `DatasetType` are very similar; the only difference is that
`DatasetType` should be used as the root container for a dataset. They behave
like ordered Python dictionaries::

    >>> list(dataset.s.keys())
    ['foo', 'bar', 'foobar']

Slicing these datasets with a list of keywords yields a `StructureType`
or `DatasetType` with only a subset of the children::

    >>> dataset.s['foo', 'foobar']
    <StructureType with children 'foo', 'foobar'>
    >>> list(dataset.s['foo', 'foobar'].keys())
    ['foo', 'foobar']

In the same way, the ``.items()`` and ``.values()`` methods are like in python
dictionaries and they iterate over sliced values.

Selecting only one child returns the child::

    >>> dataset.s['foo']
    <BaseType with data array([0, 1, 2, 3], dtype=int32)>

A `GridType` is a special container where the first child should be an
n-dimensional `BaseType`. This children should be followed by `n` additional
vector `BaseType` objects, each one describing one of the axis of the
variable::

    >>> rain = GridType('rain')
    >>> rain['rain'] = BaseType(
    ...     'rain', np.arange(6).reshape(2, 3), dimensions=('y', 'x'))
    >>> rain['x'] = BaseType('x', np.arange(3), units='degrees_east')
    >>> rain['y'] = BaseType('y', np.arange(2), units='degrees_north')
    >>> rain.array  #doctest: +ELLIPSIS
    <BaseType with data array([[0, 1, 2],
           [3, 4, 5]])>
    >>> type(rain.maps)
    <class 'collections.OrderedDict'>
    >>> for item in rain.maps.items():
    ...     print(item)
    ('x', <BaseType with data array([0, 1, 2])>)
    ('y', <BaseType with data array([0, 1])>)

There a last special container called `SequenceType`. This data structure is
analogous to a series of records (or rows), with one column for each of its
children::

    >>> cast = SequenceType('cast')
    >>> cast['depth'] = BaseType('depth', positive='down', units='m')
    >>> cast['temperature'] = BaseType('temperature', units='K')
    >>> cast['salinity'] = BaseType('salinity', units='psu')
    >>> cast['id'] = BaseType('id')
    >>> cast.data = np.array([(10., 17., 35., '1'), (20., 15., 35., '2')],
    ...     dtype=np.dtype([('depth', np.float32), ('temperature', np.float32),
    ...     ('salinity', np.float32), ('id', np.dtype('|S1'))]))

Note that the data in this case is attributed to the `SequenceType`, and is
composed of a series of values for each of the children.  pydap `SequenceType`
obects are very flexible. Data can be accessed by iterating over the object::

    >>> for record in cast.iterdata():
    ...     print(record)
    (np.float32(10.0), np.float32(17.0), np.float32(35.0), '1')
    (np.float32(20.0), np.float32(15.0), np.float32(35.0), '2')

It is possible to select only a few variables::

    >>> for record in cast['salinity', 'depth'].iterdata():
    ...     print(record)
    (np.float32(35.0), np.float32(10.0))
    (np.float32(35.0), np.float32(20.0))

    >>> cast['temperature'].dtype
    dtype('float32')
    >>> cast['temperature'].shape
    (2,)

When sliced, it yields the underlying array:
    >>> type(cast['temperature'][-1:])
    <class 'pydap.model.BaseType'>
    >>> for record in cast['temperature'][-1:].iterdata():
    ...     print(record)
    15.0

When constrained, it yields the SequenceType:
    >>> type(cast[ cast['temperature'] < 16 ])
    <class 'pydap.model.SequenceType'>
    >>> for record in cast[ cast['temperature'] < 16 ].iterdata():
    ...     print(record)
    (np.float32(20.0), np.float32(15.0), np.float32(35.0), '2')

As mentioned earlier, it is still possible to iterate directly over data::

    >>> for record in cast[ cast['temperature'] < 16 ]:
    ...     print(record)
    (np.float32(20.0), np.float32(15.0), np.float32(35.0), '2')

But this is discouraged as this will be deprecated soon. The ``.iterdata()`` is
therefore highly recommended.
"""

import copy
import operator
import re
import warnings
from collections import OrderedDict
from collections.abc import Mapping
from functools import reduce

import numpy as np

from .lib import _quote, decode_np_strings, tree, walk

__all__ = [
    "BaseType",
    "StructureType",
    "DatasetType",
    "SequenceType",
    "GridType",
    "GroupType",
]


class DapType(object):
    """The common Opendap type.

    This is a base class, defining common methods and attributes for all other
    classes in the data model.

    """

    def __init__(self, name="nameless", attributes=None, **kwargs):
        self.name = _quote(name)
        self.attributes = attributes or {}
        self.attributes.update(kwargs)

        # Set the id to the name.
        self._id = self.name

    def __repr__(self):
        return "DapType(%s)" % ", ".join(map(repr, [self.name, self.attributes]))

    # The id.
    def _set_id(self, id):
        self._id = id

        # Update children id.
        for child in self.children():
            # pass
            child.id = "%s.%s" % (id, child.name)

    def _get_id(self):
        return self._id

    id = property(_get_id, _set_id)

    def __getattr__(self, attr):
        """Attribute shortcut.

        Data classes have their attributes stored in the `attributes`
        attribute, a dictionary. For convenience, access to attributes can be
        shortcut by accessing the attributes directly::

            >>> var = DapType('var')
            >>> var.attributes['foo'] = 'bar'
            >>> var.foo
            'bar'

        This will return the value stored under `attributes`.

        """
        try:
            return self.attributes[attr]
        except (KeyError, TypeError):
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self), attr)
            )

    def children(self):
        """Return iterator over children."""
        return ()


class BaseType(DapType):
    """A thin wrapper over Numpy arrays."""

    def __init__(
        self, name="nameless", data=None, dimensions=None, attributes=None, **kwargs
    ):
        super(BaseType, self).__init__(name, attributes, **kwargs)
        self.data = data
        self.dimensions = dimensions or ()

        # these are set when not data is present (eg, when parsing a DDS)
        self._dtype = None
        self._shape = ()
        self._itemsize = None
        self._nbytes = None

    def __repr__(self):
        return "<%s with data %s>" % (type(self).__name__, repr(self.data))

    @property
    def path(self):
        return self.data.path

    @property
    def dtype(self):
        """Property that returns the data dtype."""
        return self.data.dtype

    @property
    def shape(self):
        """Property that returns the data shape."""
        return self.data.shape

    def reshape(self, *args):
        """Method that reshapes the data:"""
        self.data = self.data.reshape(*args)
        return self

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def size(self):
        return int(np.prod(self.shape))

    @property
    def itemsize(self):
        return np.asarray([], dtype=self.data.dtype).itemsize

    @property
    def nbytes(self):
        return self.itemsize * self.size

    def __copy__(self):
        """A lightweight copy of the variable.

        This will return a new object, with a copy of the attributes,
        dimensions, same name, and a view of the data.

        """
        out = type(self)(
            self.name, self.data, self.dimensions[:], self.attributes.copy()
        )
        out.id = self.id
        return out

    # Comparisons are passed to the data.
    def __eq__(self, other):
        return self.data == other

    def __ne__(self, other):
        return self.data != other

    def __ge__(self, other):
        return self.data >= other

    def __le__(self, other):
        return self.data <= other

    def __gt__(self, other):
        return self.data > other

    def __lt__(self, other):
        return self.data < other

    # Implement the sequence and iter protocols.
    def __getitem__(self, index):
        out = copy.copy(self)
        out.data = self._get_data_index(index)
        if type(self.data).__name__ == "BaseProxyDap4":
            out.attributes["checksum"] = self.data.checksum
            out.attributes["Maps"] = self.Maps
        return out

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        if self._is_string_dtype:
            for item in self.data:
                yield np.vectorize(decode_np_strings)(item)
        else:
            for item in self.data:
                yield item

    @property
    def _is_string_dtype(self):
        return hasattr(self._data, "dtype") and self._data.dtype.char == "S"

    def iterdata(self):
        """This method was added to mimic new SequenceType method."""
        return iter(self)

    def __array__(self):
        return self._get_data_index()

    def _get_data_index(self, index=Ellipsis):
        if self._is_string_dtype and isinstance(self._data, np.ndarray):
            return np.vectorize(decode_np_strings)(self._data[index])
        else:
            return self._data[index]

    def _get_data(self):
        return self._data

    def _set_data(self, data):
        self._data = data
        if np.isscalar(data):
            # Convert scalar data to
            # numpy scalar, otherwise
            # ``.dtype`` and ``.shape``
            # methods will fail.
            self._data = np.array(data)

    data = property(_get_data, _set_data)


class StructureType(DapType, Mapping):
    """A dict-like object holding other variables."""

    def __init__(self, name="nameless", attributes=None, **kwargs):
        super(StructureType, self).__init__(name, attributes, **kwargs)

        # allow some keys to be hidden:
        self._visible_keys = []
        self._dict = OrderedDict()

    def __repr__(self):
        return "<%s with children %s>" % (
            type(self).__name__,
            ", ".join(map(repr, self._visible_keys)),
        )

    def __getattr__(self, attr):
        """Lazy shortcut return children."""
        try:
            return self[attr]
        except Exception:
            return DapType.__getattr__(self, attr)

    def __contains__(self, key):
        return key in self._visible_keys

    # __iter__, __getitem__, __len__ are required for Mapping
    # From these, keys, items, values, get, __eq__,
    # and __ne__ are obtained.
    def __iter__(self):
        for key in self._dict.keys():
            if key in self._visible_keys:
                yield key

    def _all_keys(self):
        # used in ..handlers.lib
        return iter(self._dict.keys())

    def _getitem_string(self, key):
        """Assume that key is a string type"""
        try:
            return self._dict[_quote(key)]
        except KeyError:
            splitted = key.split(".")
            if len(splitted) > 1:
                try:
                    return self[splitted[0]][".".join(splitted[1:])]
                except (KeyError, IndexError):
                    return self[".".join(splitted[1:])]
            else:
                raise

    def _getitem_string_tuple(self, key):
        """Assume that key is a tuple of strings"""
        out = type(self)(self.name, data=self.data, attributes=self.attributes.copy())
        for name in key:
            out[name] = copy.copy(self._getitem_string(name))
        return out

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._getitem_string(key)
        elif isinstance(key, tuple) and all(isinstance(name, str) for name in key):
            out = copy.copy(self)
            out._visible_keys = list(key)
            return out
        else:
            raise KeyError(key)

    def __len__(self):
        return len(self._visible_keys)

    def children(self):
        # children method always yields an
        # iterator on visible children:
        for key in self._visible_keys:
            yield self[key]

    def tree(self):
        return tree(self)

    def __setitem__(self, key, item):
        key = _quote(key)
        if key != item.name:
            raise KeyError(
                'Key "%s" is different from variable name "%s"!' % (key, item.name)
            )

        if key in self:
            del self[key]
        self._dict[key] = item
        # By default added keys are visible:
        self._visible_keys.append(key)

        # Set item id.
        item.id = "%s.%s" % (self.id, item.name)

    def __delitem__(self, key):
        del self._dict[key]
        try:
            self._visible_keys.remove(key)
        except ValueError:
            pass

    def _get_data(self):
        return [var.data for var in self.children()]

    def _set_data(self, data):
        for col, var in zip(data, self.children()):
            var.data = col

    data = property(_get_data, _set_data)

    def __shallowcopy__(self):
        out = type(self)(self.name, self.attributes.copy())
        out.id = self.id
        return out

    def __copy__(self):
        """Return a lightweight copy of the Structure.

        The method will return a new Structure with cloned children, but any
        data object are not copied.

        """
        out = self.__shallowcopy__()

        # Clone all children too.
        for child in self._dict.values():
            out[child.name] = copy.copy(child)
        return out

    @property
    def type(self):
        return "Structure"

    def structures(self):
        out = {}
        Gs = [key for key in self.children() if isinstance(key, StructureType)]
        Strs = [key for key in Gs if key.type == "Structure"]
        for var in Strs:
            out.update({var.name: [key.name for key in var.children()]})
        return out

    def groups(self):
        """Returns fqn for all (nested) groups"""
        out = {}
        for var in walk(self, GroupType):
            if var.type == "Group":
                out.update({var.name: var.path})
        return out

    def sequences(self):
        out = {}
        for var in walk(self, SequenceType):
            if var.type == "Sequence":
                out.update({var.name: [key.name for key in var.children()]})
        return out

    def variables(self):
        out = {}
        Bcs = [key for key in self.children() if isinstance(key, BaseType)]
        for var in Bcs:
            out.update({var.name: var.dtype})
        return out


class DatasetType(StructureType):
    """A root Dataset.

    The Dataset is a Structure, but it names does not compose the id hierarchy:

        >>> dataset = DatasetType("A")
        >>> dataset["B"] = BaseType("B")
        >>> dataset["B"].id
        'B'
    """

    def __setitem__(self, key, item):
        # key a path-like only in DAP4
        split_by = " "
        if len(self.groups()) > 0:
            if key[0] == "/":
                key = key[1:]
            split_by += "/"
        if len(self.sequences()) > 0 or len(self.structures()) > 0:
            split_by += "."
        parts = re.split("[" + split_by + "]", key)
        N = len(parts)
        if N > 1:
            # add parent container type if not there
            if parts[0] not in self._dict:
                self._visible_keys.append(parts[0])
            #  iterate over all groups to reach DAP object
            current = self._dict
            for j in range(N - 1):
                if parts[j] not in current:  # and Flat is not None
                    #     # This current approach works when parsing a DMR
                    #     # with only Groups and arrays. Need to enable
                    #     # Sequences and Structures. This works with all
                    #     # DAP4 when creating Dataset manally.
                    current[parts[j]] = GroupType(parts[j])
                current = current[parts[j]]
            current[parts[-1]] = item
        else:
            if key[0] == "/" and len(self.groups()) > 0:
                key = key[1:]
            key = _quote(key)
            if key != item.name:
                raise KeyError(
                    'Key "%s" is different from variable name "%s"!' % (key, item.name)
                )

            if key in self:
                del self[key]

            self._dict[key] = item
            # By default added keys are visible:
            self._visible_keys.append(key)
        key = key.replace("%2E", ".")
        if len(key.split(".")) == 1:
            # The parent name does not go into the children ids.
            item.id = item.name
        else:
            parts = key.split("/")[-1]
            item.id = (".").join(parts.split("."))
            # Set item id.
            # item.id  = "%s.%s" % (parent_name, item.name)

    def _getitem_string(self, key):
        """Assume that key is a string type"""
        try:
            return self._dict[_quote(key)]
        except KeyError:
            parts = key.split("/")
            Np = len(parts)
            if Np <= 2 and set(parts) == set([""]):
                return self
            elif Np > 1 and set(parts) != set([""]):
                if parts[0] == "":
                    parts = parts[1:]
                    Np = len(parts)
                current = self
                if Np == 1:
                    return self[parts[0]]
                else:
                    for j in range(Np):
                        key_c = ("/").join(parts[j:])
                        splitted = key_c.split(".")
                        if key_c in current.keys():
                            return current[key_c]
                        elif len(splitted) > 1 and splitted[0] in current.keys():
                            return current[key_c]
                        else:
                            current = current[parts[j]]
            else:
                splitted = key.split(".")
                if len(splitted) > 1:
                    try:
                        return self[splitted[0]][".".join(splitted[1:])]
                    except (KeyError, IndexError):
                        return self[".".join(splitted[1:])]
                else:
                    raise

    def _set_id(self, id):
        """The dataset name is not included in the children ids."""
        self._id = id

        for child in self.children():
            child.id = child.name

    def to_netcdf(self, *args, **kwargs):
        try:
            from .apis.netcdf4 import NetCDF

            return NetCDF(self, *args, **kwargs)
        except ImportError:
            raise NotImplementedError(".to_netcdf requires the netCDF4 " "package.")

    def change_order(self, order):
        self._dict = OrderedDict((k, self._dict[k]) for k in order)

    @property
    def type(self):
        return ()

    @property
    def nbytes(self):
        """Returns the number of bytes occupied in aggregation by all
        uncompressed, BaseType data in the DatasetType. Attributes are
        not included in the computation of value"""
        nbytes = 0
        for var in walk(self, BaseType):
            nbytes += var.nbytes
        return nbytes

    def createDapType(self, daptype, name, **attrs):
        # Creates a temporal quasi FQN by replacing all `.` with `/`
        # since these are safe to escape. This catches all cases
        # much more cleanly. User specifies FQN but DAPtype is defined
        # within the method so OK.
        split_by = " "
        if len(self.groups()) > 0:  # true
            split_by += "/"
        else:
            if name[0] == "/":
                name = name[1:]
        if len(self.sequences()) > 0 or len(self.structures()) > 0:
            split_by += "."
        parts = re.split("[" + split_by + "]", name)
        item = daptype(name=parts[-1], **attrs)
        DatasetType.__setitem__(self, name, item)

    def createGroup(self, name, **attrs):
        """
        Creates a Group from `root`, even in the presence of nested Groups.
        Uses the `Fully Qualifying Name`. Parent `Group` must exist.

        Also: https://docs.opendap.org/index.php/DAP4:_Specification_Volume_1
        """
        path = "/"
        if name[0] != "/":
            name = "/" + name
        if len(name.split("/")) > 2:
            path = ("/").join(name.split("/")[:-1]) + "/"
        parts = re.split(r"[/]", name)
        try:
            for i in range(1, len(parts)):
                self[("/").join(parts[:i])]
        except KeyError:
            warnings.warn(
                """Failed to create `{}` because parent `{}` does not exist!
                """.format(
                    parts[-1], parts[-2]
                )
            )
            return None

        return self.createDapType(GroupType, name, path=path, **attrs)

    def createVariable(self, name, **attrs):
        """
        Creates a Variable (`BaseType`) from `root`, even in the presence of nested
        DAPTypes (i.e. `GroupType`, `SequenceType`, `StructureType`). Uses the
        `Fully Qualifying Name`. Parent `DAPType` must exist!

        Also: https://docs.opendap.org/index.php/DAP4:_Specification_Volume_1

        """
        return self.createDapType(BaseType, name, **attrs)

    def createSequence(self, name, **attrs):
        """
        Creates a Sequence (`SequenceType`) from `root`, even in the presence of nested
        DAPTypes (i.e. `GroupType`, `SequenceType`, `StructureType`).
        Uses the `Fully Qualifying Name`. Parent `DAPType` must exist!

        Also: https://docs.opendap.org/index.php/DAP4:_Specification_Volume_1
        """
        return self.createDapType(SequenceType, name, **attrs)

    def createStructure(self, name, **attrs):
        """
        Creates a Structure (`StructureType`) from `root`, even in the presence of
        nested DAPTypes (i.e. `GroupType`, `SequenceType`, `StructureType`).
        Uses the `Fully Qualifying Name`. Parent `DAPType` must exist!

        Also: https://docs.opendap.org/index.php/DAP4:_Specification_Volume_1
        """
        return self.createDapType(StructureType, name, **attrs)


class SequenceType(StructureType):
    """A container that stores data in a Numpy array.

    Here's a standard dataset for testing sequential data:

        >>> import numpy as np
        >>> data = np.array([
        ... (10, 15.2, 'Diamond_St'),
        ... (11, 13.1, 'Blacktail_Loop'),
        ... (12, 13.3, 'Platinum_St'),
        ... (13, 12.1, 'Kodiak_Trail')],
        ... dtype=np.dtype([
        ... ('index', np.int32), ('temperature', np.float32),
        ... ('site', np.dtype('|S14'))]))
        ...
        >>> seq = SequenceType('example')
        >>> seq['index'] = BaseType('index')
        >>> seq['temperature'] = BaseType('temperature')
        >>> seq['site'] = BaseType('site')
        >>> seq.data = data

    Iteraring over the sequence returns data:

        >>> for line in seq.iterdata():
        ...     print(line)
        (np.int32(10), np.float32(15.2), 'Diamond_St')
        (np.int32(11), np.float32(13.1), 'Blacktail_Loop')
        (np.int32(12), np.float32(13.3), 'Platinum_St')
        (np.int32(13), np.float32(12.1), 'Kodiak_Trail')

    The order of the variables can be changed:

        >>> for line in seq['temperature', 'site', 'index'].iterdata():
        ...     print(line)
        (np.float32(15.2), 'Diamond_St', np.int32(10))
        (np.float32(13.1), 'Blacktail_Loop', np.int32(11))
        (np.float32(13.3), 'Platinum_St', np.int32(12))
        (np.float32(12.1), 'Kodiak_Trail', np.int32(13))

    We can iterate over children:

        >>> for line in seq['temperature'].iterdata():
        ...     print(line)
        15.2
        13.1
        13.3
        12.1

    We can filter the data:

        >>> for line in seq[ seq.index > 10 ].iterdata():
        ...     print(line)
        (np.int32(11), np.float32(13.1), 'Blacktail_Loop')
        (np.int32(12), np.float32(13.3), 'Platinum_St')
        (np.int32(13), np.float32(12.1), 'Kodiak_Trail')

        >>> for line in seq[ seq.index > 10 ]['site'].iterdata():
        ...     print(line)
        Blacktail_Loop
        Platinum_St
        Kodiak_Trail

        >>> for line in (seq['site', 'temperature'][seq.index > 10]
        ...              .iterdata()):
        ...     print(line)
        ('Blacktail_Loop', np.float32(13.1))
        ('Platinum_St', np.float32(13.3))
        ('Kodiak_Trail', np.float32(12.1))

    Or slice it:

        >>> for line in seq[::2].iterdata():
        ...     print(line)
        (np.int32(10), np.float32(15.2), 'Diamond_St')
        (np.int32(12), np.float32(13.3), 'Platinum_St')

        >>> for line in seq[ seq.index > 10 ][::2]['site'].iterdata():
        ...     print(line)
        Blacktail_Loop
        Kodiak_Trail

        >>> for line in seq[ seq.index > 10 ]['site'][::2]:
        ...     print(line)
        Blacktail_Loop
        Kodiak_Trail

    """

    def __init__(self, name="nameless", data=None, attributes=None, **kwargs):
        super(SequenceType, self).__init__(name, attributes, **kwargs)
        self._data = data

    def _set_data(self, data):
        self._data = data
        for child in self.children():
            tokens = child.id[len(self.id) + 1 :].split(".")
            child.data = reduce(operator.getitem, [data] + tokens)

    def _get_data(self):
        return self._data

    data = property(_get_data, _set_data)

    def iterdata(self):
        for line in self.data:
            yield tuple(map(decode_np_strings, line))

    def __iter__(self):
        return self.iterdata()

    def __len__(self):
        return len(self.data)

    def items(self):
        # This method should be removed in pydap 3.4
        for key in self._visible_keys:
            yield (key, self[key])

    def values(self):
        # This method should be removed in pydap 3.4
        for key in self._visible_keys:
            yield self[key]

    def keys(self):
        # This method should be removed in pydap 3.4
        return iter(self._visible_keys)

    def __contains__(self, key):
        # This method should be removed in pydap 3.4
        return key in self._visible_keys

    def __getitem__(self, key):
        # If key is a string, return child with the corresponding data.
        if isinstance(key, str):
            return self._getitem_string(key)

        # If it's a tuple, return a new `SequenceType` with selected children.
        elif isinstance(key, tuple):
            out = self._getitem_string_tuple(key)
            # copy.copy() is necessary here because a view will be returned in
            # the future:
            out.data = copy.copy(self.data[list(key)])
            return out

        # Else return a new `SequenceType` with the data sliced.
        else:
            out = copy.copy(self)
            out.data = self.data[key]
            return out

    def __shallowcopy__(self):
        out = type(self)(self.name, self.data, self.attributes.copy())
        out.id = self.id
        return out

    @property
    def type(self):
        return "Sequence"


class GridType(StructureType):
    """A Grid container.

    The Grid is a Structure with an array and the corresponding axes.

    """

    def __init__(self, name="nameless", attributes=None, **kwargs):
        super(GridType, self).__init__(name, attributes, **kwargs)
        self._output_grid = True

    def __repr__(self):
        return "<%s with array %s and maps %s>" % (
            type(self).__name__,
            repr(list(self.keys())[0]),
            ", ".join(map(repr, list(self.keys())[1:])),
        )

    def __getitem__(self, key):
        # Return a child.
        if isinstance(key, str):
            return self._getitem_string(key)

        # Return a new `GridType` with part of the data.
        elif isinstance(key, tuple) and all(isinstance(name, str) for name in key):
            out = self._getitem_string_tuple(key)
            for var in out.children():
                var.data = self[var.name].data
            return out
        else:
            if not self.output_grid:
                return self.array[key]

            if not isinstance(key, tuple):
                key = (key,)

            out = copy.copy(self)
            for var, slice_ in zip(out.children(), [key] + list(key)):
                if type(self.data).__name__ == "BaseProxyDap4":
                    pass
                var.data = self[var.name].data[slice_]
            return out

    @property
    def dtype(self):
        """Return the first children dtype."""
        return self.array.dtype

    @property
    def shape(self):
        """Return the first children shape."""
        return self.array.shape

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def size(self):
        return int(np.prod(self.shape))

    @property
    def output_grid(self):
        return self._output_grid

    def set_output_grid(self, key):
        self._output_grid = bool(key)

    @property
    def array(self):
        """Return the first children."""
        return self[list(self.keys())[0]]

    def __array__(self):
        return self.array.data

    @property
    def maps(self):
        """Return the axes in an ordered dict."""
        return OrderedDict([(k, self[k]) for k in self.keys()][1:])

    @property
    def dimensions(self):
        """Return the name of the axes."""
        return tuple(list(self.keys())[1:])

    @property
    def type(self):
        return "Grid"


class GroupType(StructureType):
    """A Group container.

    A folder-like DAP container which may have other DAP types like
    Sequences, Structures, DataArrays, and other Groups, as children.

    Groups in OPeNDAP:
        - There is at least the root Group `/`.
        - Must have a name.
        - There is a single parent to a Group (tree hierarchy).
        - Attributes are Global at the Group level.
        - Can have share dimensions, which are scoped to all children (not to parent)

    Examples:

    """

    def __setitem__(self, key, item):
        StructureType.__setitem__(self, key, item)

        # The Group name does (not) go into the children ids.

        item.id = item.name
        # self._dict[key] = item

        # # # By default added keys are visible:
        # self._visible_keys.append(key)

    def _set_id(self, id):
        """The dataset name is (not) included in the children ids."""
        self._id = id

        for child in self.children():
            child.id = child.name

    @property
    def type(self):
        return "Group"



"""The ASCII response.

The ASCII response is an unnoficial response used to return the data as ASCII.
pydap's implementation is reverse engineered from the official server.

"""

try:
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

import copy

import numpy as np

from ..lib import __version__, encode
from ..model import BaseType, SequenceType, StructureType
from .dds import dds
from .lib import BaseResponse


class ASCIIResponse(BaseResponse):
    """The ASCII response."""

    __version__ = __version__

    def __init__(self, dataset):
        BaseResponse.__init__(self, dataset)
        self.headers.extend(
            [
                ("Content-description", "dods_ascii"),
                ("Content-type", "text/plain; charset=ascii"),
            ]
        )

    def __iter__(self):
        for line in dds(self.dataset):
            yield line.encode("ascii")
        yield (45 * "-" + "\n").encode("ascii")

        for line in ascii(self.dataset):
            yield line.encode("ascii")


@singledispatch
def ascii(var, printname=True):
    """A single dispatcher for the ASCII response."""
    raise StopIteration


@ascii.register(SequenceType)
def _sequenctype(var, printname=True):
    yield ", ".join([child.id for child in var.children()])
    yield "\n"
    for rec in var.iterdata():
        out = copy.copy(var)
        out.__class__ = StructureType
        out.data = rec
        for i, line in enumerate(ascii(out, printname=False)):
            line = line.strip()
            if line and i > 0:
                yield ", "
            yield line
        yield "\n"


@ascii.register(StructureType)
def _structuretype(var, printname=True):
    for child in var.children():
        for line in ascii(child, printname):
            yield line
        yield "\n"


@ascii.register(BaseType)
def _basetype(var, printname=True):
    if printname:
        yield var.id
        yield "\n"

    if not getattr(var, "shape", ()):
        yield encode(var.data)
    else:
        for indexes, value in zip(np.ndindex(var.shape), var.data.flat):
            yield "{indexes} {value}\n".format(
                indexes="[" + "][".join([str(idx) for idx in indexes]) + "]",
                value=encode(value),
            )



import os

import numpy

import pydap.client


def load_dap(file_path):
    abs_path = os.path.join(os.path.dirname(__file__), file_path)
    dataset = pydap.client.open_dap_file(abs_path)
    return dataset


def test_jpl1():
    file_path = (
        "data/daps/20220102090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1_subset.dap"
    )
    dataset = load_dap(file_path)
    values = dataset["sea_ice_fraction"].data[0, 0, 0:2]
    expected = numpy.array([98, 98], dtype="int8")
    assert numpy.array_equal(values, expected)


def test_jpl2():
    file_path = (
        "data/daps/20220531090000-JPL-L4_GHRSST-SSTfnd-MUR-GLOB-v02.0-fv04.1.dap"
    )
    dataset = load_dap(file_path)
    values = dataset["lon"].data[0:5]
    expected = numpy.array(
        [-179.99, -179.98, -179.97, -179.96, -179.95], dtype="float32"
    )
    assert numpy.array_equal(values, expected)


def test_coads():
    file_path = "data/daps/coads_climatology.nc.dap"
    dataset = load_dap(file_path)
    values = dataset["SST"][0, 2, 0:3].data
    expected = numpy.array([0.12833333, -0.05000002, -0.06363636], dtype="float32")
    numpy.testing.assert_almost_equal(values, expected)


def test_my1qnd1():
    fname = "data/daps/MY1DQND1.sst.ADD2005001.040.2006011070802.hdf.dap"
    load_dap(fname)


if __name__ == "__main__":
    pass
    # test_my1qnd1()
    # test_coads()



"""An implementation of server-side functions.

pydap implements DAP server-side functions throught a custom WSGI middleware.
This simplifies writing custom handlers, since they don't need to parse and
apply the function calls themselves.

"""

import ast
import operator
import re
from functools import reduce
from importlib.metadata import entry_points

import numpy as np
from webob import Request

from ..exceptions import ServerError
from ..handlers.lib import BaseHandler, apply_projection
from ..lib import fix_shorthand, walk
from ..model import DatasetType, SequenceType
from ..parsers import parse_ce

FUNCTION = re.compile(r"([^(]*)\((.*)\)")
RELOP = re.compile(r"(<=|<|>=|>|=~|=|!=)")


def load_functions():
    """Load all available functions from the system, returning a dictionary."""
    # Relative import of functions:
    eps = entry_points(group="pydap.function")
    Rs = [r for r in eps if r.module[:5] == "pydap"]
    nRs = [r for r in eps if r.module[:5] != "pydap"]
    base_dict = dict((r.name, r.load()) for r in Rs)

    opts_dict = dict((r.name, r.load()) for r in nRs)
    base_dict.update(opts_dict)
    return base_dict


class ServerSideFunctions(object):
    """A WebOb based middleware for handling server-side function calls.

    The middleware works by removing function calls from the request,
    forwarding the request to pydap, and then applying the functions calls to
    the returned dataset.

    """

    def __init__(self, app, **kwargs):
        self.app = app

        self.functions = load_functions()
        self.functions.update(kwargs)

    def __call__(self, environ, start_response):
        # specify that we want the parsed dataset
        environ["x-wsgiorg.want_parsed_response"] = True
        req = Request(environ)
        projection, selection = parse_ce(req.query_string)

        # check if there are any functions calls in the request
        called = any(s for s in selection if FUNCTION.match(s)) or any(
            p for p in projection if isinstance(p, str)
        )

        # ignore DAS requests and requests without functions
        path, response = req.path.rsplit(".", 1)
        if response == "das" or not called:
            return self.app(environ, start_response)

        # apply selection without any function calls
        req.query_string = "&".join(s for s in selection if not FUNCTION.match(s))
        res = req.get_response(self.app)

        # get the dataset
        method = getattr(res.app_iter, "x_wsgiorg_parsed_response", False)
        if not method:
            raise ServerError("Unable to call server-side function!")
        dataset = method(DatasetType)

        # apply selection containing server-side functions
        selection = (s for s in selection if FUNCTION.match(s))
        for expr in selection:
            if RELOP.search(expr):
                call, op, other = RELOP.split(expr)
                op = {
                    "<": operator.lt,
                    ">": operator.gt,
                    "!=": operator.ne,
                    "=": operator.eq,
                    ">=": operator.ge,
                    "<=": operator.le,
                    "=~": lambda a, b: re.match(b, a),
                }[op]
                other = ast.literal_eval(other)
            else:
                call, op, other = expr, operator.eq, 1

            # evaluate the function call
            sequence = eval_function(dataset, call, self.functions)

            # is this an inplace call?
            for var in walk(dataset, SequenceType):
                if sequence is var:
                    break
            else:
                # get the data from the resulting variable, and use it to
                # constrain the original dataset
                child = list(sequence.children())[0]
                data = np.fromiter(child.data, child.dtype)
                if data.dtype.char == "S":
                    valid = np.array(
                        list(map(lambda v: op(str(v), str(other)), data)), bool
                    )
                else:
                    valid = op(data, other)

                for sequence in walk(dataset, SequenceType):
                    sequence.data = np.rec.fromrecords(
                        [tuple(row) for row in sequence.iterdata()],
                        names=list(sequence.keys()),
                    )[valid]

        # now apply projection
        if projection:
            projection = fix_shorthand(projection, dataset)
            base = [p for p in projection if not isinstance(p, str)]
            func = [p for p in projection if isinstance(p, str)]

            # apply non-function projection
            out = apply_projection(base, dataset)

            # apply function projection
            for call in func:
                var = eval_function(dataset, call, self.functions)
                for child in walk(var):
                    parent = reduce(operator.getitem, [out] + child.id.split(".")[:-1])
                    if child.name not in parent.keys():
                        parent[child.name] = child
                        break
            dataset = out

        # Return the original response (DDS, DAS, etc.)
        path, response = req.path.rsplit(".", 1)
        res = BaseHandler.responses[response](dataset)

        return res(environ, start_response)


def eval_function(dataset, function, functions):
    """Evaluate a given function on a dataset.

    This function parses and evaluates a (possibly nested) function call,
    returning its result.

    """
    name, args = FUNCTION.match(function).groups()

    def tokenize(input):
        start = pos = count = 0
        for char in input:
            if char == "(":
                count += 1
            elif char == ")":
                count -= 1
            elif char == "," and count == 0:
                yield input[start:pos]
                start = pos + 1
            pos += 1
        yield input[start:]

    def parse(token):
        if FUNCTION.match(token):
            return eval_function(dataset, token, functions)
        else:
            try:
                names = re.sub(r"\[.*?\]", "", str(token)).split(".")
                return reduce(operator.getitem, [dataset] + names)
            except Exception:
                try:
                    return ast.literal_eval(token)
                except Exception:
                    return token

    args = map(parse, tokenize(args))
    func = functions[name]

    return func(dataset, *args)



import pytest

from pydap.client import open_url

base_url = "dap4://test.opendap.org"


def test_coads():
    url = base_url + "/opendap/hyrax/data/nc/coads_climatology.nc"
    pydap_ds = open_url(url)
    pydap_ds["COADSX"][10:12:1]


def test_groups():
    url = base_url + ":8080/opendap/dmrpp_test_files/"
    pydap_ds = open_url(url + "ATL03_20181228015957_13810110_003_01.2var.h5.dmrpp")
    pydap_ds["/gt1r/bckgrd_atlas/bckgrd_int_height"][0:10]


@pytest.mark.skip(reason="bug in testserver")
def test_maps():
    url = base_url + ":8080/opendap/hyrax/data/nc/coads_climatology.nc"
    pydap_ds = open_url(url)
    data = pydap_ds["SST"][0:2:1, 40:42:1, 1:10:1]
    print(data.array[:].data)


if __name__ == "__main__":
    test_maps()



from urllib.parse import quote_plus

from . import get_cookies


def setup_session(
    openid, password, username=None, check_url=None, session=None, verify=False
):
    """
    A special call to get_cookies.setup_session that is tailored for
    ESGF credentials.

    username should only be necessary for a CEDA openid.
    """
    session = get_cookies.setup_session(
        _uri(openid),
        username=username,
        password=password,
        check_url=check_url,
        session=session,
        verify=verify,
    )
    # Connections can be kept alive on the ESGF:
    session.headers.update([("Connection", "keep-alive")])
    return session


def _uri(openid):
    """
    Create ESGF authentication url.
    This function might be sensitive to a
    future evolution of the ESGF security.
    """

    def generate_url(dest_url):
        dest_node = _get_node(dest_url)

        try:
            url = (
                dest_node + "/esg-orp/j_spring_openid_security_check.htm?"
                "openid_identifier=" + quote_plus(openid)
            )
        except TypeError:
            raise UserWarning("OPENID not set. ESGF connection failed.")
        if _get_node(openid) == "https://ceda.ac.uk":
            return [url, None]
        else:
            return url

    return generate_url


def _get_node(url):
    return "/".join(url.split("/")[:3]).replace("http:", "https:")



import logging

from netCDF4 import Dataset

from ..handlers.lib import BaseHandler
from ..server.devel import LocalTestServer

_logger = logging.getLogger(__name__)

_private_atts = ["server"]


class NetCDF(Dataset):
    def __init__(self, dataset, *args, **kwargs):
        self.server = LocalTestServer(
            application=BaseHandler(dataset=dataset), as_process=True
        )
        self.server.start()
        super(NetCDF, self).__init__(self.server.url, *args, **kwargs)

    def __setattr__(self, name, value):
        if name in _private_atts:
            self.__dict__[name] = value
        else:
            super(Dataset, self).__setattr__(name, value)

    def __getattr__(self, name):
        if name in _private_atts:
            return self.__dict__[name]
        else:
            return super(Dataset, self).__getattr__(name)

    def __delattr__(self, name):
        if name not in _private_atts:
            super(Dataset, self).__delattr__(name)

    def close(self):
        # Shutdown server on close:
        self.server.shutdown()

        # Close but without checking for error:
        Dataset._close(self, False)



"""Test the version response."""

import unittest

from webtest import TestApp as App

from pydap.handlers.lib import BaseHandler
from pydap.lib import __version__
from pydap.tests.datasets import VerySimpleSequence


class TestVersionResponse(unittest.TestCase):
    """Simple tests for the version response.

    Here we don't test the actual response body, since it will vary from
    system to system and breaks continuous integration.

    """

    def setUp(self):
        """Create a simple WSGI app with a dataset."""
        app = App(BaseHandler(VerySimpleSequence))
        self.res = app.get("/.ver")

    def test_status(self):
        """Test the status code."""
        self.assertEqual(self.res.status, "200 OK")

    def test_content_type(self):
        """Test that content-type is correctly set."""
        self.assertEqual(self.res.content_type, "application/json")

    def test_charset(self):
        """Test that charset is correctly set."""
        self.assertEqual(self.res.charset, "utf-8")

    def test_headers(self):
        """Test that all headers are present."""
        self.assertEqual(
            self.res.headers["Content-Type"], "application/json; charset=utf-8"
        )
        self.assertEqual(self.res.headers["Content-description"], "dods_version")
        self.assertEqual(self.res.headers["OPeNDAP-Server"], "pydap/" + __version__)



"""A handler for remote datasets.

DAP handlers convert from different data formats (NetCDF, eg) to the internal
pydap model. The pydap client is just a handler that converts from a remote
dataset to the internal model.

"""

import copy
import gzip
import io

# handlers should be set by the application
# http://docs.python.org/2/howto/logging.html#configuring-logging-for-a-library
import logging
import pprint
import re
import sys
import warnings as _warnings
from io import BytesIO
from itertools import chain

import numpy
from requests.utils import urlparse, urlunparse

from pydap.handlers.lib import BaseHandler, ConstraintExpression, IterData
from pydap.lib import (
    DAP2_ARRAY_LENGTH_NUMPY_TYPE,
    DEFAULT_TIMEOUT,
    START_OF_SEQUENCE,
    BytesReader,
    StreamReader,
    _quote,
    combine_slices,
    encode,
    fix_slice,
    hyperslab,
    walk,
)
from pydap.model import BaseType, GridType, SequenceType, StructureType
from pydap.net import GET, raise_for_status
from pydap.parsers import parse_ce
from pydap.parsers.das import add_attributes, parse_das
from pydap.parsers.dds import dds_to_dataset
from pydap.parsers.dmr import dmr_to_dataset
from pydap.responses.dods import DAP2_response_dtypemap

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

BLOCKSIZE = 512


class DAPHandler(BaseHandler):
    """Build a dataset from a DAP base URL."""

    def __init__(
        self,
        url,
        application=None,
        session=None,
        output_grid=True,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_charset="ascii",
        protocol=None,
    ):

        self.application = application
        self.session = session
        self.output_grid = output_grid
        self.timeout = timeout
        self.verify = verify
        self.user_charset = user_charset
        self.url = url

        # urlparse returns an additional var compared to
        # urlsplit: `param`. Will toss it.
        scheme, netloc, path, _, query, fragment = urlparse(self.url)
        self.scheme = scheme
        self.netloc = netloc
        self.path = path
        self.query = query
        self.fragment = fragment

        self.protocol = self.determine_protocol(protocol)

        self.projection, self.selection = parse_ce(self.query, self.protocol)
        arg = (
            self.scheme,
            self.netloc,
            self.path,
            "",
            "&".join(self.selection),
            self.fragment,
        )
        self.base_url = urlunparse(arg)
        self.make_dataset()
        self.add_proxies()

    def determine_protocol(self, protocol):
        if protocol == "dap4":
            self.scheme = "http"
            return protocol
        elif protocol == "dap2":
            return protocol
        elif self.scheme == "dap4":
            self.scheme = "http"
            return "dap4"
        elif self.query[:4] == "dap4":
            self.sheme = "http"
            return "dap4"
        else:
            extension = self.path.split(".")[-1]
            if extension in ["dmr", "dap"]:
                return "dap4"
            elif extension in ["dds", "dods"]:
                return "dap2"
            else:
                _warnings.warn(
                    "PyDAP was unable to determine the DAP protocol "
                    "defaulting to DAP2 which is consider legacy and "
                    "may result in slower responses. For more, see "
                    "go to https://www.opendap.org/faq-page."
                )
                return "dap2"

    def make_dataset(
        self,
    ):
        if self.protocol == "dap4":
            self.dataset_from_dap4()
        else:
            self.dataset_from_dap2()
            self.attach_das()

    def dataset_from_dap4(self):
        dmr_url = urlunparse(
            (
                self.scheme,
                self.netloc,
                self.path + ".dmr",
                "",
                _quote(self.query),
                self.fragment,
            )
        )
        r = GET(
            dmr_url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        raise_for_status(r)
        dmr = safe_charset_text(r, self.user_charset)
        self.dataset = dmr_to_dataset(dmr)

    def dataset_from_dap2(self):
        # escape for certain characters
        dds_url = urlunparse(
            (
                self.scheme,
                self.netloc,
                self.path + ".dds",
                "",
                _quote(self.query),
                self.fragment,
            )
        )
        r = GET(
            dds_url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        raise_for_status(r)
        dds = safe_charset_text(r, self.user_charset)
        self.dataset = dds_to_dataset(dds)

    def attach_das(self):
        # Also pull the DAS and add additional attributes
        das_url = urlunparse(
            (
                self.scheme,
                self.netloc,
                self.path + ".das",
                "",
                self.query,
                self.fragment,
            )
        )
        r = GET(
            das_url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        raise_for_status(r)
        das = safe_charset_text(r, self.user_charset)
        add_attributes(self.dataset, parse_das(das))

    def add_proxies(self):
        if self.protocol == "dap4":
            self.add_dap4_proxies()
        else:
            self.add_dap2_proxies()

    def add_dap4_proxies(self):
        # remove any projection from the base_url, leaving selections
        for var in walk(self.dataset, BaseType):
            if var.path is not None:
                var_name = var.path + "/" + var.name
            else:
                var_name = var.name
            var.data = BaseProxyDap4(
                self.base_url,
                var_name,
                var.dtype,
                var.shape,
                application=self.application,
                session=self.session,
                timeout=self.timeout,
                verify=self.verify,
            )

        # apply projections to BaseType only
        # CE for sequences and structs
        # are not ready (see https://github.com/pydap/pydap/issues/314)
        for var in self.projection:
            target = self.dataset
            while var:
                token, index = var.pop(0)
                target = target[token]
                if isinstance(target, BaseType):
                    target.data.slice = fix_slice(index, target.shape)

    def add_dap2_proxies(self):
        # now add data proxies
        for var in walk(self.dataset, BaseType):
            var.data = BaseProxyDap2(
                self.base_url,
                var.id,
                var.dtype,
                var.shape,
                application=self.application,
                session=self.session,
                timeout=self.timeout,
                verify=self.verify,
            )
        for var in walk(self.dataset, SequenceType):
            template = copy.copy(var)
            var.data = SequenceProxy(
                self.base_url,
                template,
                selection=self.selection,
                application=self.application,
                session=self.session,
                timeout=self.timeout,
                verify=self.verify,
            )

        # apply projections
        for var in self.projection:
            target = self.dataset
            while var:
                token, index = var.pop(0)
                target = target[token]
                if isinstance(target, BaseType):
                    target.data.slice = fix_slice(index, target.shape)
                elif isinstance(target, GridType):
                    index = fix_slice(index, target.array.shape)
                    target.array.data.slice = index
                    for s, child in zip(index, target.maps):
                        target[child].data.slice = (s,)
                elif isinstance(target, SequenceType):
                    target.data.slice = index

        # retrieve only main variable for grid types:
        for var in walk(self.dataset, GridType):
            var.set_output_grid(self.output_grid)


def get_charset(r, user_charset):
    charset = r.charset
    if not charset:
        charset = user_charset
    return charset


def safe_charset_text(r, user_charset):
    if r.content_encoding == "gzip":
        return (
            gzip.GzipFile(fileobj=BytesIO(r.body))
            .read()
            .decode(get_charset(r, user_charset))
        )
    else:
        r.charset = get_charset(r, user_charset)
        return r.text


def safe_dds_and_data(r, user_charset):
    if r.content_encoding == "gzip":
        raw = gzip.GzipFile(fileobj=BytesIO(r.body)).read()
    else:
        raw = r.body
    dds, data = raw.split(b"\nData:\n", 1)
    return dds.decode(get_charset(r, user_charset)), data


def safe_dmr_and_data(r, user_charset):
    if r.content_encoding == "gzip":
        raw = gzip.GzipFile(fileobj=BytesIO(r.body)).read()
    else:
        raw = r.body
    logger.info("Saving and splitting dmr+")
    try:
        dmr, data = raw.split(b"</Dataset>", 1)
    except ValueError:
        logger.exception("Failed to split the following DMR+ \n %s" % raw)
        import codecs
        import pickle

        picked_response = str(codecs.encode(pickle.dumps(r), "base64").decode())
        header = "pickled response (base64)"
        logger.exception(
            header
            + ": \n ----BEGIN PICKLE----- \n %s \n -----END PICKLE-----"
            % picked_response
        )

    dmr = dmr[4:] + b"</Dataset>"
    dmr = dmr.decode(get_charset(r, user_charset))
    data = data[3:]
    return dmr, data


class BaseProxyDap2(object):
    """A proxy for remote base types.

    This class behaves like a Numpy array, proxying the data from a base type
    on a remote dataset.
    """

    def __init__(
        self,
        baseurl,
        id,
        dtype,
        shape,
        slice_=None,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_charset="ascii",
    ):
        self.baseurl = baseurl
        self.id = id
        self.dtype = dtype
        self.shape = shape
        self.slice = slice_ or tuple(slice(None) for s in self.shape)
        self.application = application
        self.session = session
        self.timeout = timeout
        self.verify = verify
        self.user_charset = user_charset

    def __repr__(self):
        return "BaseProxy(%s)" % ", ".join(
            map(repr, [self.baseurl, self.id, self.dtype, self.shape, self.slice])
        )

    def __getitem__(self, index):
        # build download url

        index = combine_slices(self.slice, fix_slice(index, self.shape))
        scheme, netloc, path, params, query, fragment = urlparse(self.baseurl)
        url = urlunparse(
            (
                scheme,
                netloc,
                path + ".dods",
                "",
                self.id + hyperslab(index) + "&" + _quote(query),
                fragment,
            )
        ).rstrip("&")

        # download and unpack data
        logger.info("Fetching URL: %s" % url)
        r = GET(
            url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )

        raise_for_status(r)
        dds, data = safe_dds_and_data(r, self.user_charset)

        # Parse received dataset:
        dataset = dds_to_dataset(dds)
        dataset.data = unpack_dap2_data(BytesReader(data), dataset)
        return dataset[self.id].data

    def __len__(self):
        return self.shape[0]

    def __iter__(self):
        return iter(self[:])

    # Comparisons return a boolean array
    def __eq__(self, other):
        return self[:] == other

    def __ne__(self, other):
        return self[:] != other

    def __ge__(self, other):
        return self[:] >= other

    def __le__(self, other):
        return self[:] <= other

    def __gt__(self, other):
        return self[:] > other

    def __lt__(self, other):
        return self[:] < other


class BaseProxyDap4(BaseProxyDap2):

    def __init__(
        self,
        baseurl,
        id,
        dtype,
        shape,
        slice_=None,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
        user_charset="ascii",
    ):
        self.baseurl = baseurl
        self.id = id
        self.dtype = dtype
        self.shape = shape
        self.slice = slice_ or tuple(slice(None) for s in self.shape)
        self.application = application
        self.session = session
        self.timeout = timeout
        self.verify = verify
        self.user_charset = user_charset

    def __repr__(self):
        return "Dap4BaseProxy(%s)" % ", ".join(
            map(repr, [self.baseurl, self.id, self.dtype, self.shape, self.slice])
        )

    def __getitem__(self, index):
        # build download url
        index = combine_slices(self.slice, fix_slice(index, self.shape))
        scheme, netloc, path, _, query, fragment = urlparse(self.baseurl)
        ce = "dap4.ce=" + self.id + hyperslab(index)
        url = urlunparse((scheme, netloc, path + ".dap", "", ce, fragment)).rstrip("&")

        # download and unpack data
        logger.info("Fetching URL: %s" % url)

        r = GET(
            url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )

        raise_for_status(r)
        dmr, data = safe_dmr_and_data(r, self.user_charset)

        # Parse received dataset:
        dataset = dmr_to_dataset(dmr)
        dataset = unpack_dap4_data(BytesReader(data), dataset)

        self.checksum = dataset[self.id].attributes["checksum"]
        self.data = dataset[self.id].data
        return self.data


class SequenceProxy(object):
    """A proxy for remote sequences.

    This class behaves like a Numpy structured array, proxying the data from a
    sequence on a remote dataset. The data is streamed from the dataset,
    meaning it can be treated one record at a time before the whole data is
    downloaded.

    """

    shape = ()

    def __init__(
        self,
        baseurl,
        template,
        selection=None,
        slice_=None,
        application=None,
        session=None,
        timeout=DEFAULT_TIMEOUT,
        verify=True,
    ):
        self.baseurl = baseurl
        self.template = template
        self.selection = selection or []
        self.slice = slice_ or (slice(None),)
        self.application = application
        self.session = session
        self.timeout = timeout
        self.verify = verify

        # this variable is true when only a subset of the children are selected
        self.sub_children = False

    @property
    def dtype(self):
        return self.template.dtype

    def __repr__(self):
        return "SequenceProxy(%s)" % ", ".join(
            map(repr, [self.baseurl, self.template, self.selection, self.slice])
        )

    def __copy__(self):
        """Return a lightweight copy of the object."""
        return self.__class__(
            self.baseurl,
            self.template,
            self.selection[:],
            self.slice[:],
            self.application,
        )

    def __getitem__(self, key):
        """Return a new object representing a subset of the data."""
        out = copy.copy(self)

        # return the data for a children
        if isinstance(key, str):
            out.template = out.template[key]

        # return a new object with requested columns
        elif isinstance(key, list):
            out.sub_children = True
            out.template._visible_keys = key

        # return a copy with the added constraints
        elif isinstance(key, ConstraintExpression):
            out.selection.extend(str(key).split("&"))

        # slice data
        else:
            if isinstance(key, int):
                key = slice(key, key + 1)
            out.slice = combine_slices(self.slice, (key,))

        return out

    @property
    def url(self):
        """Return url from where data is fetched."""
        scheme, netloc, path, params, query, fragment = urlparse(self.baseurl)
        url = urlunparse(
            (
                scheme,
                netloc,
                path + ".dods",
                "",
                self.id + hyperslab(self.slice) + "&" + "&".join(self.selection),
                fragment,
            )
        ).rstrip("&")

        return url

    @property
    def id(self):
        """Return the id of this sequence."""
        if self.sub_children:
            id_ = ",".join(child.id for child in self.template.children())
        else:
            id_ = self.template.id
        return id_

    def __iter__(self):
        # download and unpack data
        r = GET(
            self.url,
            self.application,
            self.session,
            timeout=self.timeout,
            verify=self.verify,
        )
        raise_for_status(r)

        i = r.app_iter
        if not hasattr(i, "__next__"):
            i = iter(i)

        # Fast forward past the DDS header
        # the pattern could span chunk boundaries though so make sure to check
        pattern = b"Data:\n"
        last_chunk = find_pattern_in_string_iter(pattern, i)

        if last_chunk is None:
            raise ValueError(
                "Could not find data segment in response from {}".format(self.url)
            )

        # Then construct a stream consisting of everything from
        # 'Data:\n' to the end of the chunk + the rest of the stream
        def stream_start():
            yield last_chunk

        stream = StreamReader(chain(stream_start(), i))

        return unpack_sequence(stream, self.template)

    def __eq__(self, other):
        return ConstraintExpression("%s=%s" % (self.id, encode(other)))

    def __ne__(self, other):
        return ConstraintExpression("%s!=%s" % (self.id, encode(other)))

    def __ge__(self, other):
        return ConstraintExpression("%s>=%s" % (self.id, encode(other)))

    def __le__(self, other):
        return ConstraintExpression("%s<=%s" % (self.id, encode(other)))

    def __gt__(self, other):
        return ConstraintExpression("%s>%s" % (self.id, encode(other)))

    def __lt__(self, other):
        return ConstraintExpression("%s<%s" % (self.id, encode(other)))


def unpack_sequence(stream, template):
    """Unpack data from a sequence, yielding records."""
    # is this a sequence or a base type?
    sequence = isinstance(template, SequenceType)

    # if there are no children, we use the template as the only column
    cols = list(template.children()) or [template]

    # if there are no strings and no nested sequences we can unpack record by
    # record easily
    simple = all(isinstance(c, BaseType) and c.dtype.char not in "SU" for c in cols)

    if simple:
        dtype = numpy.dtype([("", c.dtype, c.shape) for c in cols])
        marker = stream.read(4)
        while marker == START_OF_SEQUENCE:
            rec = numpy.frombuffer(stream.read(dtype.itemsize), dtype=dtype)[0]
            if not sequence:
                rec = rec[0]
            yield rec
            marker = stream.read(4)
    else:
        marker = stream.read(4)
        while marker == START_OF_SEQUENCE:
            rec = unpack_children(stream, template)
            if not sequence:
                rec = rec[0]
            else:
                rec = tuple(rec)
            yield rec
            marker = stream.read(4)


def unpack_children(stream, template):
    """Unpack children from a structure, returning their data."""
    cols = list(template.children()) or [template]

    out = []
    for col in cols:
        # sequences and other structures
        if isinstance(col, SequenceType):
            out.append(IterData(list(unpack_sequence(stream, col)), col))
        elif isinstance(col, StructureType):
            out.append(tuple(unpack_children(stream, col)))

        # unpack arrays
        else:
            out.extend(convert_stream_to_list(stream, col.dtype, col.shape, col.id))
    return out


def convert_stream_to_list(stream, parser_dtype, shape, id):
    out = []
    response_dtype = DAP2_response_dtypemap(parser_dtype)
    if shape:
        n = numpy.frombuffer(stream.read(4), DAP2_ARRAY_LENGTH_NUMPY_TYPE)[0]
        count = response_dtype.itemsize * n
        if response_dtype.char in "S":
            # Consider on 'S' and not 'SU' because
            # response_dtype.char should never be
            data = []
            for _ in range(n):
                k = numpy.frombuffer(stream.read(4), DAP2_ARRAY_LENGTH_NUMPY_TYPE)[0]
                data.append(stream.read(k))
                stream.read(-k % 4)
            out.append(
                numpy.array([str(x.decode("ascii")) for x in data], "S").reshape(shape)
            )
        else:
            stream.read(4)  # read additional length
            try:
                out.append(
                    numpy.frombuffer(stream.read(count), response_dtype)
                    .astype(parser_dtype)
                    .reshape(shape)
                )
            except ValueError as e:
                if str(e) == "total size of new array must be unchanged":
                    # server-side failure.
                    # it is expected that the user should be mindful of this:
                    raise RuntimeError(
                        (
                            "variable {0} could not be properly "
                            "retrieved. To avoid this "
                            "error consider using open_url(..., "
                            "output_grid=False)."
                        ).format(id)
                    )
                else:
                    raise
            if response_dtype.char == "B":
                # Unsigned Byte type is packed to multiples of 4 bytes:
                stream.read(-n % 4)

    # special types: strings and bytes
    elif response_dtype.char in "S":
        # Consider on 'S' and not 'SU' because
        # response_dtype.char should never be
        # 'U'
        k = numpy.frombuffer(stream.read(4), DAP2_ARRAY_LENGTH_NUMPY_TYPE)[0]
        out.append(str(stream.read(k).decode("ascii")))
        stream.read(-k % 4)
    # usual data
    else:
        out.append(
            numpy.frombuffer(
                stream.read(response_dtype.itemsize), response_dtype
            ).astype(parser_dtype)[0]
        )
        if response_dtype.char == "B":
            # Unsigned Byte type is packed to multiples of 4 bytes:
            stream.read(3)
    return out


def unpack_dap2_data(xdr_stream, dataset):
    """Unpack a string of encoded data, returning data as lists."""
    return unpack_children(xdr_stream, dataset)


def decode_chunktype(chunk_type):
    encoding = "{0:03b}".format(chunk_type)
    if sys.byteorder == "little":
        # If our machine's byteorder is little,
        # we need to swap since the chunk_type is always big endian
        encoding = encoding[::-1]
    last_chunk = bool(int(encoding[0]))
    error = bool(int(encoding[1]))
    endian = {"0": ">", "1": "<"}[encoding[2]]
    return last_chunk, error, endian


def get_count(variable):
    count = int(numpy.array(variable.shape).prod())
    item_size = numpy.dtype(variable.dtype).itemsize
    count = count * item_size
    return count


def decode_variable(buffer, start, stop, variable, endian):
    dtype = variable.dtype
    dtype = dtype.newbyteorder(endian)
    data = numpy.frombuffer(buffer[start:stop], dtype=dtype).astype(dtype)
    data = data.reshape(variable.shape)
    return data


def stream2bytearray(xdr_stream):
    last = False
    buffer = bytearray()
    while not last:
        chunk = numpy.frombuffer(xdr_stream.read(4), dtype=">u4")
        chunk_size = (chunk & 0x00FFFFFF)[0]
        chunk_type = ((chunk >> 24) & 0xFF)[0]
        last, error, endian = decode_chunktype(chunk_type)
        buffer.extend(xdr_stream.read(chunk_size))
    return buffer


def get_endianness(xdr_stream):
    chunk_header = xdr_stream.peek(4)[0:4]
    chunk_header = numpy.frombuffer(chunk_header, dtype=">u4")[0]
    chunk_type = (chunk_header >> 24) & 0xFF
    last, error, endian = decode_chunktype(chunk_type)
    return endian


def unpack_dap4_data(xdr_stream, dataset):
    endian = get_endianness(xdr_stream)
    checksum_dtype = numpy.dtype(endian + "u4")
    buffer = stream2bytearray(xdr_stream)

    start = 0
    for variable in walk(dataset, BaseType):
        # variable_name = variable.name
        # variable = dataset[variable_name]
        count = get_count(variable)
        stop = start + count
        data = decode_variable(
            buffer, start=start, stop=stop, variable=variable, endian=endian
        )
        checksum = numpy.frombuffer(
            buffer[stop : stop + 4], dtype=checksum_dtype
        ).byteswap("=")
        if isinstance(variable, BaseType):
            variable._set_data(data)
        elif isinstance(variable, GridType):
            variable._set_data([data.data])
        variable.attributes["checksum"] = checksum
        # Jump over the 4 byte chunk_header
        start = stop + 4
    return dataset


def find_pattern_in_string_iter(pattern, i):
    last_chunk = b""
    length = len(pattern)
    for this_chunk in i:
        last_chunk += this_chunk
        m = re.search(pattern, last_chunk)
        if m:
            return last_chunk[m.end() :]
        last_chunk = last_chunk[-length:]


def dump():  # pragma: no cover
    """Unpack dods response into lists.

    Return pretty-printed data.

    """
    dods = sys.stdin.read()
    dds, xdrdata = dods.split(b"\nData:\n", 1)
    dataset = dds_to_dataset(dds)
    xdr_stream = io.BytesIO(xdrdata)
    data = unpack_dap2_data(xdr_stream, dataset)
    pprint.pprint(data)
